#######################################################
#
#Copyright (C) 2011 by Elliot Borenstein
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.
#
########################################################
#
# eRig Auto-Rigging Utility Prototype
# Created by Elliot Borenstein
# V1.2 10/2011
#
# Usage: Import the module (import eRig) and call eRig.openWindow()
# eRig requires Pymel 1.0 or later to work.
# eRig has been tested on Maya 2010 and 2011.
# Documentation is available from imgd.elliotb.com.
#
#
# Should you feel the desire to look through the code,
# please keep in mind that this was a rapid prototype
# and thus the code isn't exactly clean or well laid out.
#
#
# If you have any questions or comments, please email
# me at Elliot@ElliotB.com.
#
#
#######################################################
#
#
# Known Issues/Future Plans:
#
# -If the scene is set to units other than centimeters (the default), pivots will be in the wrong place
#    -Something to do with an API function somewhere assuming centimeters
# -Anything tagged with "//*TODO"
#
#
#######################################################
#
#
# Change Log:
#
# 1.0 - Initial Release
# 1.1 - 12/17/10 - added skinning utils to the utils tab
# 1.2 - 10/09/11 - Fixed IK twist issue on arms/legs
#                     -Code assumed that the user's current ik solver was RP in some cases.
#						It is now explicit about what solvers to use in all cases.
#
#
#
#######################################################


"""

Rig Creation Overview

When the button in the UI is pressed to build the rig, the buildRig() function is called.

First, this function steps through the skeleton, identifying individual components (arm, spine, etc),
and building the joint hierarchies needed for them based on the positions the user has defined.
While this is happening, objects subclassed from ModuleContainerInterface are created with references to
the new joints.

Next, the new joints have their orientations corrected, and the postOrient() method of each
container object from the last step is called. This method generally cleans up any spare joints
used for proper orientation, or corrects orientations as need.

Next, the skeleton is duplicated; The original will be used for the rig, while the duplicate
will be the one on which the mesh is bound and to which finalized animations are baked. The bind
skeleton is parent constrained to the rig skeleton.

Next, the controllers for the rig skeleton are created. Once again, the containers are itterated over,
calling the buildControllers() method. Any controller groups created by these method calls are stored,
and are then retrieved in the next step where they are parented under a single group.

Finally, if requested, proxy geometry is created by iterating over all joints in the bind skeleton, and
is constrained to the bind skeleton.


"""

import pymel.core as pc
import pymel.core.datatypes as pcd
from functools import partial

#enums for curve types
kCircleCurve = 0
kBoxCurve = 1
kSpatulaCurve = 2
kSphereCurve = 3
kHCurve = 4
kCogCurve = 5
kMasterCurve = 6

#enums for sides of the body
kSideCenter = 0
kSideLeft = 1
kSideRight = 2
kSideNone = 3

#Prefixes for sides of the body
kSideLeftPrefix = "L_"
kSideRightPrefix = "R_"
kSideCenterPrefix = ""
kSideNonePrefix = ""

#Attribute names
kFkJointAttr = "eFkJointAttr"
kIkJointAttr = "eIkJointAttr"

#Postfixes for IK and FK controls/joints
kIkPrefix = "_IK"
kFkPrefix = "_FK"

#Misc limb names
kSpineName = "Spine"
kTorsoRootName = "Torso_Root"
kCogName = "COG"
kNeckName = "Neck"
kHeadName = "Head"
kHeadTipName = "Head_Tip"
kJawName = "Jaw"
kEyeName = "Eye"
kHipsName = "Hips"
kHipsTipName = "HipsTip"
kClavicleName = "Clavicle"
kJawName = "Jaw"
kJawTipName = "JawTip"

#arm limb Names
kShoulderName = "Shoulder"
kElbowName = "Elbow"
kForearmName = "Forearm"
kWristName = "Wrist"
kHandName = "Hand"

#leg limb names
kFemurName = "Femur"
kKneeName = "Knee"
kAnkleName = "Ankle"
kBallName = "Ball"
kToeName = "Toe"
kHeelName = "Heel"

#attributes for rig configuration
kCogAlignXAttr = "eAlignX_Attr" #is the character aligned down the x or the z?
kCogCreateProxyAttr = "eCreateProxyAttr" #should proxy geo be created?
kArmFkIkAttr = "eArmFkIkAttr" #what modes should the arms have (ik/fk/both)
kLegFkIkAttr = "eLegFkIkAttr" #what modes should the legs have
kFkOnly = 0
kIkOnly = 1
kFkIk = 2
kEyeVisitedAttr = "eEyeVisitedAttr" #added to an eye joint when the builder has visited it

#mirroring constants
kMirrorXY = 0
kMirrorYZ = 1
kMirrorXZ = 2


class globalData(object):
	"""
	passed along to each module builder as the rig is being built
	Stores various bits of global data about the rig

	"""
	def __init__(self):
		self.buildList = []
		self.frontX = False #True if the rig faces down +X, False if +Z
		self.createProxy = True #True if proxy geometry should be made. Else false
	
	def append(self, item):
		"""append item to the build list"""
		self.buildList.append(item)
	
	def getBuildList(self):
		"""Returns the build list"""
		return self.buildList
	
	def setFrontX(self, frontX):
		"""sets the value of frontX (true if the model faces down +X, false if +Z)"""
		self.frontX = frontX
		
	def getFrontX(self):
		"""Returns the value of frontX"""
		return self.frontX
	
	def setCreateProxy(self, createProxy):
		"""Sets whether of not proxy geo should be created"""
		self.createProxy = createProxy
	
	def getCreateProxy(self):
		"""Returns the proxy geometry option"""
		return self.createProxy


class ModuleContainerInterface(object):
	"""
	Base class for module containers. Each modlue (spine, arm, etc) is placed in a container
	which then handles building the rig for that module. Subclasses may re-implement whatever
	functions they need.
	
	"""
	def __init__(self):
		self.groups = []
	def buildControllers(self, rigData):
		"""Builds the controllers for the module. rigData is the rigData (of type globalData) for the rig"""
		pass
	def postOrient(self, rigData):
		"""Handles setup on the rig after the joints have been oriented"""
		pass
	def getGroups(self):
		"""Returns the list of controller groups made by this module"""
		return self.groups
	def addGroups(self, groups):
		"""Adds a controller group or list of groups to the internal goup list"""
		if isinstance(groups, list):
			#it is a list, so get each item
			for item in groups:
				self.groups.append(item)
		else:
			#just one item, so store it
			self.groups.append(groups)

class HipsContainer(ModuleContainerInterface):
	"""
	Container class for the hip joint.
	Creates an H controller along the negative of the facing axis.
	
	"""
	
	def __init__(self, hips):
		ModuleContainerInterface.__init__(self)
		#Store the hip joint
		self.hips = hips
		self.tip = None
		
		#recurse down the DAG to find the tip generated for 
		#orientation purposes
		for child in hips.getChildren():
			if pc.nodeType(child) == "joint" and child.otherType.get() == kHipsTipName:
				self.tip = child
				break
	
	def buildControllers(self, rigData):
		"""
		Makes an H controller, oriented and positioned based on the
		facing axis in the rigData
		
		"""
		if rigData.getFrontX():
			trans = [0, -2, 0]
			rot = [0, 90, 0]
		else:
			trans = [0, 0, -2]
			rot = [90, 0, 90]
		
		#attach the controller, set the color, and store the group
		ctrl, ctrlGroup = attachController(self.hips, kHCurve, rotation = rot,\
										translation = trans, name = self.hips.nodeName() + "CTRL")
		assignColorOverride(ctrl, self.hips.side.get())
		self.addGroups(ctrlGroup)
	
	def postOrient(self, rigData):
		"""
		Removes the tip joint from the hips,
		and reparents any children from the tip to the hips.
		
		"""	
		
		#unparent the tip
		pc.parent(self.tip, w = True)
		
		#The hip joint was moved toward the parent joint a little and the tip placed where it was to be
		#to allow for joint orienting. So, move the hips to the right spot
		self.hips.setTranslation(self.tip.getTranslation("world"), "world")
		
		#reparent anything under the tip to be under the hips
		for child in self.tip.getChildren():
			pc.parent(child, self.hips)
		
		#delete the tip
		pc.delete(self.tip)
		self.tip = None

class SpineContainer(ModuleContainerInterface):
	"""Container for handling a spine, which can have an arbitrary number of joints"""
	def __init__(self, spineRoot):
		ModuleContainerInterface.__init__(self)
		#Store the spine root and make an empty list for the other joints
		self.spineRoot = spineRoot
		self.spines = []
		self.torsoRoot = None
		
		#recurse down the DAG to find all the spine joints
		curNode = spineRoot
		while(curNode != None):
			self.spines.append(curNode)
			children = curNode.getChildren()
			
			if children == []:
				#no more children, so we're done
				curNode = None
			else:
				#there are children, so check if they are something to deal with
				for child in children:
					if pc.nodeType(child) == "joint":
						if child.otherType.get() == kSpineName:
							curNode = child
							break
						elif child.otherType.get() == kTorsoRootName:
							self.torsoRoot = child
							curNode = None
						else:
							curNode = None
					else:
						curNode = None
	
	def buildControllers(self, rigData):
		"""Attaches a circle controller to each spine joint"""
		for joint in self.spines:
			ctrl, ctrlGroup = attachController(joint, kCircleCurve, name = joint.nodeName() + "CTRL")
			assignColorOverride(ctrl, joint.side.get())
			self.addGroups(ctrlGroup)
	
	def postOrient(self, rigData):
		"""
		If some of the spine joints are primarilly pointed straight up,
		the default orientation with y-up will cause issues. So, we check
		if this is the case and re-orient them with z-up is needed.
		
		Also, deletes the torso_root if nothing is attached to it
		
		"""
		#gather the largest magnitudes for each axis by going through each joint
		x = 0
		y = 0
		z = 0
		for spine in self.spines:
			#get the first child, calculate vector to it
			child = spine.getChildren()
			if len(child) != 0:
				child = child [0]
				vect = child.getTranslation("world") - spine.getTranslation("world")
				vect = pcd.normal(vect)
				
				#store the normalized components and their absolute values
				vX = vect[0]
				vaX = pcd.abs(vX)
				vY = vect[1]
				vaY = pcd.abs(vY)
				vZ = vect[2]
				vaZ = pcd.abs(vZ)
				
				#now see if the new value for that axis is larger than the old one
				if vaX > pcd.abs(x):
					x = vX
				if vaY > pcd.abs(y):
					y = vY
				if vaZ > pcd.abs(z):
					z = vZ
		
		#now figure out the largest axis. Only act if it is y
		ax = pcd.abs(x)
		ay = pcd.abs(y)
		az = pcd.abs(z)
		
		if ax > ay:
			#either x or z
			pass
		elif ay > az:
			#y is largest
			#we need to re-orient
			for spine in self.spines:
				spine.orientJoint("xyz", sao = "zup", ch = False, zso = True)
		else:
			#z is largest
			pass
				
				
		
		#if there is nothing attached to the torso, delete it
		if self.torsoRoot:
			children = self.torsoRoot.getChildren()
			if children == []:
				pc.delete(self.torsoRoot)
				self.torsoRoot = None 
	
class ClavicleContainer(ModuleContainerInterface):
	"""Container for a clavicle, creating a sphere controller"""
	def __init__(self, clavicle):
		ModuleContainerInterface.__init__(self)
		
		#just store the clavicle
		self.clavicle = clavicle
	
	def buildControllers(self, rigData):
		"""Create a sphere controller for the clavicle and attach it"""
		ctrl, ctrlGroup = attachController(self.clavicle, kSphereCurve, name = self.clavicle.nodeName() + "CTRL")
		assignColorOverride(ctrl, self.clavicle.side.get())
		self.addGroups(ctrlGroup)
	
	def postOrient(self, rigData):
		#does nothing at the moment
		return
"""		
		#this stuff doesn't work because the mirrorJoint functionality isn't documented
		side = self.clavicle.side.get()
	
		#mirror right side joints
		if side == kSideRight:
			self.clavicle.jointOrientX.set(self.clavicle.jointOrientX.get() * -1)
			self.clavicle.jointOrientY.set(self.clavicle.jointOrientY.get() * -1)
			self.clavicle.jointOrientZ.set(self.clavicle.jointOrientZ.get() + 180)
"""
		
class JawContainer(ModuleContainerInterface):
	"""Container for the jaw. Creates a simple circle controller"""
	def __init__(self, jaw):
		ModuleContainerInterface.__init__(self)
		
		#store the jaw and find the tip
		self.jaw = jaw
		self.tip = None
		
		for child in jaw.getChildren():
			if pc.nodeType(child) == "joint" and child.otherType.get() == kJawTipName:
				self.tip = child
				break
	
	def buildControllers(self, rigData):
		"""Assign a circle controller to the jaw"""
		ctrl, ctrlGroup = attachController(self.jaw, kCircleCurve, name = self.jaw.nodeName() + "CTRL")
		assignColorOverride(ctrl, self.jaw.side.get())
		ctrlGroup.s.set([.4, .4, .4])
		self.addGroups(ctrlGroup)
	
	def postOrient(self, rigData):
		"""Deletes the jaw tip if nothing is attached to it"""
		if self.tip:
			children = self.tip.getChildren()
			if children == []:
				pc.delete(self.tip)
				self.tip = None 

class EyeContainer(ModuleContainerInterface):
	"""
	Container for eyes. Assigns a single control curve to handle the orientation of
	multiple eye joints.
	
	Unlike other containers, the eye is not passed to the constructor. Instead, each eye is
	added to the container through the addEye method.
	
	"""
	def __init__(self):
		ModuleContainerInterface.__init__(self)
		self.eyes = []
	
	def addEye(self, eye):
		self.eyes.append(eye)
	
	def buildControllers(self, rigData):
		"""
		Creates a circle controller at the midpoint of all
		eye joints to control their orientation. The circle is
		oriented based no the facing axis from the rigData.
		
		"""
		if rigData.getFrontX():
			rot = [0, 0, 0]
		else:
			rot = [0, 90, 0]
		
		#make a controller
		loc = circleController()
		locGroup = pc.group(loc, w = True, n = "EyeCTRL_group")
		pc.rename(loc, "EyeCTRL")
		
		assignColorOverride(loc)
		
		#calculate the average position of the eyes
		x = 0
		y = 0
		z = 0
		numEyes = len(self.eyes)
		for eye in self.eyes:
			pc.connectAttr(loc.r, eye.r)
			pos = eye.getTranslation("world")
			x += pos[0]
			y += pos[1]
			z += pos[2]
		
		locGroup.setTranslation([x / numEyes, y / numEyes, z / numEyes], "world")
		locGroup.s.set([.2, .2, .2])
		locGroup.r.set(rot)
		
		theParent = self.eyes[0].getParent()
		if(theParent):
			parentConstraint = pc.parentConstraint(theParent, locGroup, mo = True)
			parentConstraint.attr("lockOutput").set(True)
		
		#loc unused controller attributes
		pc.setAttr(loc.tx, k = False, l = True, cb = False)
		pc.setAttr(loc.ty, k = False, l = True, cb = False)
		pc.setAttr(loc.tz, k = False, l = True, cb = False)
		pc.setAttr(loc.sx, k = False, l = True, cb = False)
		pc.setAttr(loc.sy, k = False, l = True, cb = False)
		pc.setAttr(loc.sz, k = False, l = True, cb = False)
		
		self.addGroups(locGroup)

class HeadContainer(ModuleContainerInterface):
	"""Container for the head. Handles both the head and aribtrary number of necks"""
	def __init__(self, neckRoot):
		ModuleContainerInterface.__init__(self)
		#Store the neck and make an empty list
		self.neckRoot = neckRoot
		self.necks = []
		self.head = None
		self.headTip = None
		
		#recurse down the DAG to find everything
		curNode = neckRoot
		while(curNode != None):
			self.necks.append(curNode)
			children = curNode.getChildren()
			
			if children == []:
				#no more children, so we're done
				curNode = None
			else:
				for child in children:
					if pc.nodeType(child) == "joint":
						if child.otherType.get() == kNeckName:
							curNode = child
							break
						elif child.otherType.get() == kHeadName:
							self.head = child
							curNode = None
						else:
							curNode = None
					else:
						curNode = None
		
		#if we found a head, find the head tip
		if self.head:
			children = self.head.getChildren()
			for child in children:
				if pc.nodeType(child) == "joint" and child.otherType.get() == kHeadTipName:
					self.headTip = child
					break
	
	def buildControllers(self, rigData):
		"""Creates a spatula controller for each neck joint, and a circle for the head"""
		if rigData.getFrontX():
			#rot = [-90, 90, 0]
			rot = [90, 0, 0]
		else:
			#rot = [180, 0, 90]
			rot = [90, 0, 0]
		
		for joint in self.necks:
			ctrl, ctrlGroup = attachController(joint, kSpatulaCurve, rotation = rot, name = joint.nodeName() + "CTRL")
			assignColorOverride(ctrl, joint.side.get())
			self.addGroups(ctrlGroup)
		ctrl, ctrlGroup = attachController(self.head, kCircleCurve, name = self.head.nodeName() + "CTRL")
		assignColorOverride(ctrl, self.head.side.get())
		self.addGroups(ctrlGroup)
	
	def postOrient(self, rigData):
		"""
		As with the spine, reorients the neck joints if they generally point
		along the y axis.
		
		Also, deletes the head tip if nothing is attached to it
		
		"""
		
		#gather the largest magnitudes for each axis by going through each joint
		x = 0
		y = 0
		z = 0
		for neck in self.necks:
			#get the first child, calculate vector to it
			child = neck.getChildren()
			if len(child) != 0:
				child = child [0]
				vect = child.getTranslation("world") - neck.getTranslation("world")
				vect = pcd.normal(vect)
				
				vX = vect[0]
				vaX = pcd.abs(vX)
				vY = vect[1]
				vaY = pcd.abs(vY)
				vZ = vect[2]
				vaZ = pcd.abs(vZ)
				
				#now see if the new value for that axis is larger than the old one
				if vaX > pcd.abs(x):
					x = vX
				if vaY > pcd.abs(y):
					y = vY
				if vaZ > pcd.abs(z):
					z = vZ
		
		#now figure out the largest axis. Only act if it is y
		ax = pcd.abs(x)
		ay = pcd.abs(y)
		az = pcd.abs(z)
		
		if ax > ay:
			#either x or z
			pass
		elif ay > az:
			#y is largest
			#we need to re-orient
			#we want to orient with y along the opposite of the facing direction
			if rigData.getFrontX():
				sao = "xdown"
			else:
				sao = "zdown"
			for neck in self.necks:
				neck.orientJoint("xyz", sao = sao, ch = False, zso = True)
		else:
			#z is largest
			pass
		
		#if there is nothing attached to the tip, delete it
		if self.headTip:
			children = self.headTip.getChildren()
			if children == []:
				pc.delete(self.headTip)
				self.headTip = None 
		

class ArmContainer(ModuleContainerInterface):
	"""Container for the arms. Handles setup of IK, Fk, and switching where needed"""
	def __init__(self, shoulder, buildList = None, fkIk = kFkIk):
		"""
		shoulder - the shoulder joint
		buildList - a globalData object used when trying to build any rig components
					that are children of this arm's joints
		fkIk - what mode should the arm use (kFkOnly, kIkOnly, kFKIK)
		
		"""
		
		ModuleContainerInterface.__init__(self)
		#go down the hierarchy and make sure we have all arm joints
		self.shoulder = None
		self.elbow = None
		self.forearm = None
		self.wrist = None
		self.hand = None
		self.hasAllJoints = False
		self.fkIk = fkIk
		
		#lists of things that need to be parented to each joint
		#these come from componenets attached to each joint and built during traversal
		self.shoulderList = []
		self.elbowList = []
		self.forearmList = []
		self.wristList = []
		self.handList = []
		
		#make sure we have a shoulder
		if(shoulder and pc.nodeType(shoulder) == "joint" and shoulder.otherType.get() == kShoulderName):
			self.shoulder = shoulder
		else:
			pyError("No Shoulder")
			return
		
		#try to find the elbow joint.
		#Also, if there are any joints under the shoulder that are not the elbow,
		#try to build them.	
		children = self.shoulder.getChildren()
		for child in children:
			if pc.nodeType(child) == "joint" and str(child.otherType.get()) == kElbowName:
				self.elbow = child
			elif buildList != None:
				built = buildComponent(child, buildList)
				if built:
					self.shoulderList.append(built)
		if not self.elbow:
			#no forearm (which is OK), but also no elbow (which is not)
			pyError("No elbow")
			return
		
		#try to find the forearm or wrist (if there is no forearm)
		#also, build any attached components, as above
		children = self.elbow.getChildren()		
		for child in children:
			if pc.nodeType(child) == "joint" and str(child.otherType.get()) == kForearmName:
				self.forearm = child
			elif pc.nodeType(child) == "joint" and str(child.otherType.get()) == kWristName:
				self.wrist = child
			elif buildList != None:
				built = buildComponent(child, buildList)
				if built:
					self.elbowList.append(built)
		if(self.forearm):
			children = self.forearm.getChildren()
			for child in children:
				if pc.nodeType(child) == "joint" and str(child.otherType.get()) == kWristName:
					self.wrist = child
				elif buildList != None:
					built = buildComponent(child, buildList)
					if built:
						self.forearmList.append(built)
		elif not self.elbow:
			#no forearm (which is OK), but also no wrist (which is not)
			pyError("No wrist")
			return
		
		
		#try to find the hand
		#also, build any attached components, as above
		children = self.wrist.getChildren()
		for child in children:
			if pc.nodeType(child) == "joint" and str(child.otherType.get()) == kHandName:
				self.hand = child
			elif buildList != None:
				built = buildComponent(child, buildList)
				if built:
					self.wristList.append(built)
		if(not self.hand):
			pyError("No hand")
			return
		
		#build any children of the hand
		if buildList != None:
			children = self.hand.getChildren()
			for child in children:
				built = buildComponent(child, buildList)
				if built:
					self.handList.append(built)
		
		self.hasAllJoints = True #found everything
	
	def duplicate(self):
		"""
		Duplicates the arm's joint chain from the shoulder.
		Returns a new ArmContainer from that joint.
		
		"""
		
		#//*TODO - Check for, and remove, extra child joints?
		newShoulder = pc.duplicate(self.shoulder)[0]
		return ArmContainer(newShoulder)
	
	def getJointList(self):
		"""Returns a list of the arm joints in this container (missing joints will be None)"""
		list = [self.shoulder, self.elbow, self.forearm, self.wrist, self.hand]
		return list
	
	def constrainTo(self, driver):
		"""Orient-constrains the arm-joints in this container to those in the driver container"""
		driverList = driver.getJointList()
		drivenList = self.getJointList()
		for i in range(len(drivenList)):
			if(driverList[i] and drivenList[i]):
				constraint = pc.orientConstraint(driverList[i], drivenList[i], mo = False)
				constraint.attr("lockOutput").set(True)
				constraint.attr("interpType").set(2) #shortest interpolation type
	
	def buildControllers(self, rigData):
		"""
		Sets up the controllers for the arm based on the original
		ik/fk option passed to the constructor.
		
		If both IK and FK are selected, creates a switch at the hand to blend the
		joint constraints and to toggle visibility of the controls.
		
		"""
		
		#create IK and FK arms as needed
		#//*TODO - don't need to duplicate and constrain the arm joints if only one or the other
		if self.fkIk == kFkOnly or self.fkIk == kFkIk:
			fkArm = buildFkArm(self)
			self.addGroups(fkArm.getGroups())
		if self.fkIk == kIkOnly or self.fkIk == kFkIk:
			ikArm = buildIkArm(self)
			self.addGroups(ikArm.getGroups())
		
		#check if we need a switch, and make one if so.
		if self.fkIk == kFkIk:
			#determine the name for the switch control
			side = self.shoulder.side.get()
			sidePrefix = getSidePrefix(side)
			switchName = sidePrefix + "HandCTRL"
			
			#get a curve
			switch = getController(kSpatulaCurve)
			switchGroup = pc.group(switch, w = True, n = switchName + "_group")
			
			assignColorOverride(switch, side)
			
			#align and constrain the control
			#Pymel 1.0.0 (Maya 2011 default) bugs out on setMatrix
			#so do it the old fashioned way
			#switchGroup.setMatrix(self.hand.getMatrix(ws=True), ws=True) #align group to object
			pc.xform(switchGroup, m = self.hand.getMatrix(ws=True), ws=True) #align group to object
			
			pc.rename(switch, switchName)
			switch.rx.set(90) #make it point in the right direction
			switch.ry.set(90) #make it point in the right direction
			pc.makeIdentity(switch, apply = True, r = True) #freeze the transforms
			pc.parentConstraint(self.hand, switchGroup, mo = True)
			
			#setup the IK/FK switch
			#hide all transform attributes and setup driven keys
			switch.addAttr("ikFkSwitch", k = True, min = 0, max = 10, dv = 10)
			#I can't believe there's no easier way to lock and hide
			#all of the attributes at once than to do the following...
			pc.setAttr(switch.tx, k = False, l = True, cb = False)
			pc.setAttr(switch.ty, k = False, l = True, cb = False)
			pc.setAttr(switch.tz, k = False, l = True, cb = False)
			pc.setAttr(switch.rx, k = False, l = True, cb = False)
			pc.setAttr(switch.ry, k = False, l = True, cb = False)
			pc.setAttr(switch.rz, k = False, l = True, cb = False)
			pc.setAttr(switch.sx, k = False, l = True, cb = False)
			pc.setAttr(switch.sy, k = False, l = True, cb = False)
			pc.setAttr(switch.sz, k = False, l = True, cb = False)
			pc.setAttr(switch.v, k = False, l = True, cb = False)
			pc.setDrivenKeyframe([fkArm.shoulderCtrl.v, fkArm.elbowCtrl.v, fkArm.handCtrl.v], cd = switch.ikFkSwitch, dv = 0, v = 0, ott = "step")
			pc.setDrivenKeyframe([fkArm.shoulderCtrl.v, fkArm.elbowCtrl.v, fkArm.handCtrl.v], cd = switch.ikFkSwitch, dv = 5, v = 1)
			
			pc.setDrivenKeyframe(ikArm.handCtrl.v, cd = switch.ikFkSwitch, dv = 0, v = 1, ott = "step")
			pc.setDrivenKeyframe(ikArm.handCtrl.v, cd = switch.ikFkSwitch, dv = 5, v = 0)
			
			#now, go through each joint and grab it's orient constraint
			#so we can set the driven keys for which joint chain controls the arm
			for baseJoint in self.getJointList():
				for child in baseJoint.getChildren():
					if pc.nodeType(child) == "orientConstraint":
						#get a list of it's targets and associated weight attributes
						targets = child.getTargetList()
						weights = child.getWeightAliasList()
						numTargets = len(targets)
						
						#for each target, check if it's Ik or Fk, and set a driven key appropriately
						#when creating the joints, they are given attributes to signify ik or fk
						#we use those here to figure out which is which
						for i in range(numTargets):
							if targets[i].hasAttr(kFkJointAttr):
								pc.setDrivenKeyframe(weights[i], cd = switch.ikFkSwitch, dv = 0, v = 0, ott = "linear")
								pc.setDrivenKeyframe(weights[i], cd = switch.ikFkSwitch, dv = 10, v = 1)
							elif targets[i].hasAttr(kIkJointAttr):
								pc.setDrivenKeyframe(weights[i], cd = switch.ikFkSwitch, dv = 0, v = 1, ott = "linear")
								pc.setDrivenKeyframe(weights[i], cd = switch.ikFkSwitch, dv = 10, v = 0)
			
			#store the controller group
			self.addGroups(switchGroup)
	
	def postOrient(self, rigData):
		"""Does nothing"""
		return
	
	"""
		#this stuff doesn't work because the mirrorJoint function isn't documented
		
		side = self.shoulder.side.get()
	
		#mirror right side joints
		if side == kSideRight:
			for joint in self.getJointList():
				if joint != self.hand:
					joint.jointOrientX.set(joint.jointOrientX.get() * -1)
					joint.jointOrientY.set(joint.jointOrientY.get() * -1)
					
			#now adjust the shoulder if we have no clavicle
			clav = self.shoulder.getParent()
			if not (clav and pc.nodeType(clav) == "joint" and clav.otherType.get() == kClavicleName):
				self.shoulder.jointOrientZ.set(self.shoulder.jointOrientZ.get() + 180)
	"""
		

class LegContainer(ModuleContainerInterface):
	"""Container for the legs. Handles setup of IK, Fk, and switching where needed"""
	def __init__(self, femur, buildList = None, fkIk = kIkOnly):
		"""
		femur - the femur joint
		buildList - a globalData object used when trying to build any rig components
					that are children of this leg's joints
		fkIk - what mode should the arm use (kFkOnly, kIkOnly, kFKIK)
		
		Works very similar to the ArmContainer
		
		"""
		
		ModuleContainerInterface.__init__(self)
		#go down the hierarchy and make sure we have all arm joints
		self.femur = None
		self.knee = None
		self.ankle = None
		self.ball = None
		self.toe = None
		self.hasAllJoints = False
		
		self.fkIk = fkIk
		
		#lists of things that need to be parented to each joint
		#these come from componenets attached to each joint and built during traversal
		self.femurList = []
		self.kneeList = []
		self.ankleList = []
		self.ballList = []
		self.toeList = []
		
		#For each of the joints in the leg, go through its children
		#We are looking for either the next leg joint, or children corresponding to connected rig parts
		#that need to be built (in that case, we send them to be built before going on).
		#If this is a duplicate leg (IK or FK), buildList will be None and the children won't be built
		#If we fail to find any joint, just return, as self.hasAllJoints will be false
		if(femur and pc.nodeType(femur) == "joint" and str(femur.otherType.get()) == kFemurName):
			self.femur = femur
		else:
			return
			
		children = self.femur.getChildren()
		for child in children:
			if pc.nodeType(child) == "joint" and str(child.otherType.get()) == kKneeName:
				self.knee = child
			elif buildList != None:
				built = buildComponent(child, buildList)
				if built:
					self.femurList.append(built)
		if(not self.knee):
			return
		
		children = self.knee.getChildren()
		for child in children:
			if pc.nodeType(child) == "joint" and str(child.otherType.get()) == kAnkleName:
				self.ankle = child
			elif buildList != None:
				built = buildComponent(child, buildList)
				if built:
					self.kneeList.append(built)
		if(not self.ankle):
			return
		
		children = self.ankle.getChildren()
		for child in children:
			if pc.nodeType(child) == "joint" and str(child.otherType.get()) == kBallName:
				self.ball = child
			elif buildList != None:
				built = buildComponent(child, buildList)
				if built:
					self.ankleList.append(built)
		if(not self.ball):
			return
		
		children = self.ball.getChildren()
		for child in children:
			if pc.nodeType(child) == "joint" and str(child.otherType.get()) == kToeName:
				self.toe = child
			elif buildList != None:
				built = buildComponent(child, buildList)
				if built:
					self.ballList.append(built)
		if(not self.toe):
			return
		
		#no more leg joints to find, but check for children of the toe
		if buildList:
				children = self.toe.getChildren()
				for child in children:
					built = buildComponent(child, buildList)
					if built:
						self.toeList.append(built)
		
		self.hasAllJoints = True #found everything
	
	def duplicate(self):
		"""Duoplicates the leg from the femur down. Returns a new LegContainer for the duplicate"""
		newFemur = pc.duplicate(self.femur)[0]
		return LegContainer(newFemur)
	
	def getJointList(self):
		"""Returns the joints for the leg as a list"""
		list = [self.femur, self.knee, self.ankle, self.ball, self.toe]
		return list
	
	def constrainTo(self, driver):
		"""
		Constrains the joints in this container to the ones in the driver legContainer.
		
		"""
		driverList = driver.getJointList()
		drivenList = self.getJointList()
		for i in range(len(drivenList)):
			if(driverList[i] and drivenList[i]):
				constraint = pc.orientConstraint(driverList[i], drivenList[i], mo = False)
				constraint.attr("lockOutput").set(True)
				constraint.attr("interpType").set(2) #shortest interpolation type
	
	def buildControllers(self, rigData):
		"""
		Sets up the controllers for the leg based on the original
		ik/fk option passed to the constructor.
		
		If both IK and FK are selected, creates a switch at the ankle to blend the
		joint constraints and to toggle visibility of the controls.
		
		"""
		
		#make the IK and FK joint chains as needed
		if self.fkIk == kFkOnly or self.fkIk == kFkIk:
			fkLeg = buildFkLeg(self)
			self.addGroups(fkLeg.getGroups())
		if self.fkIk == kIkOnly or self.fkIk == kFkIk:
			ikLeg = buildIkLeg(self, self.heelPos)
			self.addGroups(ikLeg.getGroups())
		
		#Check if we need a switch
		if self.fkIk == kFkIk:
			#figure out the front face of the rig and possibly adjust the switch control rotation
			if rigData.getFrontX():
				rot = [0, 90, 0]
			else:
				rot = [0, 90, 0]
			
			#determine the control's name
			side = self.femur.side.get()
			sidePrefix = getSidePrefix(side)
			switchName = sidePrefix + "FootCTRL"
			
			#get th curve
			switch = getController(kSpatulaCurve)
			switchGroup = pc.group(switch, w = True, n = switchName + "_group")
			
			assignColorOverride(switch, side)
			
			
			#align and constrain the control
			#Pymel 1.0.0 (Maya 2011 default) bugs out on setMatrix so use xform
			#switchGroup.setMatrix(self.ankle.getMatrix(ws=True), ws=True) #align group to object
			pc.xform(switchGroup, m = self.ankle.getMatrix(ws=True), ws=True) #align group to object
			
			pc.rename(switch, switchName)
			switch.r.set(rot) #make it point in the right direction
			pc.makeIdentity(switch, apply = True, r = True) #freeze the transforms
			pc.parentConstraint(self.ankle, switchGroup, mo = True)
			
			#setup the IK/FK switch, hide the transform attributes, setup driven keys, etc
			switch.addAttr("ikFkSwitch", k = True, min = 0, max = 10)
			#I can't believe there's no easier way to do the following...
			pc.setAttr(switch.tx, k = False, l = True, cb = False)
			pc.setAttr(switch.ty, k = False, l = True, cb = False)
			pc.setAttr(switch.tz, k = False, l = True, cb = False)
			pc.setAttr(switch.rx, k = False, l = True, cb = False)
			pc.setAttr(switch.ry, k = False, l = True, cb = False)
			pc.setAttr(switch.rz, k = False, l = True, cb = False)
			pc.setAttr(switch.sx, k = False, l = True, cb = False)
			pc.setAttr(switch.sy, k = False, l = True, cb = False)
			pc.setAttr(switch.sz, k = False, l = True, cb = False)
			pc.setAttr(switch.v, k = False, l = True, cb = False)
			pc.setDrivenKeyframe([fkLeg.femurCtrl.v, fkLeg.kneeCtrl.v,\
								fkLeg.ankleCtrl.v, fkLeg.ballCtrl.v], cd = switch.ikFkSwitch, dv = 0, v = 0, ott = "step")
			pc.setDrivenKeyframe([fkLeg.femurCtrl.v, fkLeg.kneeCtrl.v,\
								fkLeg.ankleCtrl.v, fkLeg.ballCtrl.v], cd = switch.ikFkSwitch, dv = 5, v = 1)
			
			pc.setDrivenKeyframe(ikLeg.footCtrl.v, cd = switch.ikFkSwitch, dv = 0, v = 1, ott = "step")
			pc.setDrivenKeyframe(ikLeg.footCtrl.v, cd = switch.ikFkSwitch, dv = 5, v = 0)
			
			#now, go through each joint and grab it's orient constraint
			#so we can set the driven keys for which joint chain controls the leg
			for baseJoint in self.getJointList():
				for child in baseJoint.getChildren():
					if pc.nodeType(child) == "orientConstraint":
						#get a list of it's targets and associated weight attributes
						targets = child.getTargetList()
						weights = child.getWeightAliasList()
						numTargets = len(targets)
						
						#for each target, check if it's Ik or Fk, and set a driven key appropriately
						#The special identifier attributes were added to the ik/fk joints earlier
						for i in range(numTargets):
							if targets[i].hasAttr(kFkJointAttr):
								pc.setDrivenKeyframe(weights[i], cd = switch.ikFkSwitch, dv = 0, v = 0, ott = "linear")
								pc.setDrivenKeyframe(weights[i], cd = switch.ikFkSwitch, dv = 10, v = 1)
							elif targets[i].hasAttr(kIkJointAttr):
								pc.setDrivenKeyframe(weights[i], cd = switch.ikFkSwitch, dv = 0, v = 1, ott = "linear")
								pc.setDrivenKeyframe(weights[i], cd = switch.ikFkSwitch, dv = 10, v = 0)
							
			self.addGroups(switchGroup)


def getSidePrefix(side):
	"""
	Converts a side enum (kSideLeft, kSideRight, kSideCenter, kSideNode)
	to the appropriate string prefix for that side, if any.
	
	"""
	if side == kSideLeft:
		return kSideLeftPrefix
	elif side == kSideRight:
		return kSideRightPrefix
	elif side == kSideCenter:
		return kSideCenterPrefix
	elif side == kSideNone:
		return kSideNonePrefix
	else:
		return ""

def buildComponent(component, buildList):
	"""
	build each individual component type
	
	Returns the root of whatever joint chain was just built,
	or returns None if component was not of a known compoenent type.
	
	buildList is a globalData object
	
	"""
	if component and pc.nodeType(component) == "joint":
		typeName = component.otherType.get()
		if typeName == kSpineName:
			return buildSpine(component, buildList)
		elif typeName == kNeckName:
			return buildHead(component, buildList)
		elif typeName == kShoulderName:
			return buildArm(component, buildList)
		elif typeName == kHipsName:
			return buildHips(component, buildList)
		elif typeName == kFemurName:
			return buildLeg(component, buildList)
		elif typeName == kCogName:
			return buildCog(component, buildList)
		elif typeName == kClavicleName:
			return buildClavicle(component, buildList)
		elif typeName == kEyeName:
			return buildEye(component, buildList)
		elif typeName == kJawName:
			return buildJaw(component, buildList)
	
	return None

def buildRig():
	"""
	entry point for rig creation. Select a COG joint before running
	
	Builds a rig from the selected set of components (rooted at a COG joint).
	The original components are deleted, and nothing is selected at completion.
	
	"""
	
	#Find the first cog joint in the selection
	#the rig will be built based on its children
	selection = pc.selected()
	
	cog = None
	
	for item in selection:
		if pc.nodeType(item) == "joint" and item.otherType.get() == kCogName:
			cog = item
			break
	
	if not cog:
		pyError("COG not found in selection!")
		return
	
	#create the rig data
	rigData = globalData()
	
	#build the skeleton recursively
	outCog = buildComponent(cog, rigData)
	
	#get the list of all the containers to be built
	buildList = rigData.getBuildList()
	
	#orient everything, then orient the cog to the world
	outCog.orientJoint("xyz", sao = "yup", ch = True, zso = True)
	outCog.orientJoint("none", ch = False, zso = True)
	
	#do post orient tasks on the build list
	for item in buildList:
		item.postOrient(rigData)
	
	#duplicate to get bind skeleton
	bindCog = pc.duplicate(outCog)[0]
	
	#hide the control skeleton
	pc.hide(outCog)
	
	#constrain the bind to the control
	bindConstraints = [] #list of all constraints that are created for the bind skeleton
	connectBindSkeleton(outCog, bindCog, bindConstraints)
	
	#store all the bind constraints in a group
	bindConstraintsGroup = pc.group(em = True, w = True, n = "bindConstraintsGroup")
	pc.parent(bindConstraints, bindConstraintsGroup)
	
	#build the controllers
	#first build the COG control
	cogCtrl = cogController()
	cogCtrl.setTranslation(outCog.getTranslation("world"), "world")
	assignColorOverride(cogCtrl)
	pc.makeIdentity(cogCtrl, a = True)
	pc.rename(cogCtrl, outCog.nodeName() + "CTRL")
	pc.setAttr(cogCtrl.sx, k = False, l = True, cb = False) #lock scale
	pc.setAttr(cogCtrl.sy, k = False, l = True, cb = False)
	pc.setAttr(cogCtrl.sz, k = False, l = True, cb = False)
	constraint = pc.parentConstraint(cogCtrl, outCog, mo = True) #constrain to the joint
	constraint.attr("lockOutput").set(True)
	constraint.attr("interpType").set(2) #shortest interpolation type
	
	#now build the master control
	masterCtrl = masterController()
	masterCtrl.setTranslation(outCog.getTranslation("world"), "world")
	masterCtrl.ty.set(0)
	assignColorOverride(masterCtrl)
	pc.makeIdentity(masterCtrl, a = True)
	pc.rename(masterCtrl, "MasterCTRL")
	
	#//*TODO
	#we're going to hide scale for the moment, but it could be added back in easilly
	#however, having the scale option complicates key baking
	pc.setAttr(masterCtrl.sx, k = False, l = True, cb = False)
	pc.setAttr(masterCtrl.sy, k = False, l = True, cb = False)
	pc.setAttr(masterCtrl.sz, k = False, l = True, cb = False)
	
	#make a bunch of groups
	rigGroup = pc.group(em = True, w = True, n = "RigGroup")
	doNotTouchGroup = pc.group(em = True, w = True, n = "DoNotTouch")
	scaleGroup = pc.group(em = True, w = True, n = "ScaleGroup")
	
	#setup the rig group hierarchy
	pc.parent(doNotTouchGroup, rigGroup)
	pc.parent(bindConstraintsGroup, doNotTouchGroup)
	pc.parent(masterCtrl, doNotTouchGroup)
	pc.parent(scaleGroup, masterCtrl)
	pc.parent(outCog, scaleGroup)
	pc.parent(cogCtrl, scaleGroup)
	
	#hide the transforms for the various groups
	lockAndHideTransform(doNotTouchGroup)
	lockAndHideTransform(rigGroup)
	lockAndHideTransform(scaleGroup)
	lockAndHideTransform(bindConstraintsGroup)
	
	
	#build the controlllers for everything else
	for item in buildList:
		item.buildControllers(rigData)
		
	#handle controller groups
	groups = []
	for item in buildList:
		itemGroups = item.getGroups()
		if itemGroups != []:
			groups.append(itemGroups)
	#parent all controller groups under a main group
	controllerGroup = pc.group(em = True, w = True, n = "Controllers_Group")
	for item in groups:
		pc.parent(item, controllerGroup)
		
	pc.parent(controllerGroup, scaleGroup)
	
	#make display layers
	pc.select(bindCog)
	jointLayer = pc.createDisplayLayer(n = "Joints_DL", nr = True)
	jointLayer.displayType.set(2)
	
	
	pc.select([controllerGroup, masterCtrl, cogCtrl])
	layer = pc.createDisplayLayer(n = "CTRL_DL", nr = True)
	
	#make proxy geometry if requested
	if rigData.getCreateProxy():
		lowPolys = []
		#make materials for the sides
		leftMat, leftSG = createLambert("RIG_LProxy_Material", [0, 0, 1])
		centerMat, centerSG = createLambert("RIG_CProxy_Material", [1, 1, 0])
		rightMat, rightSG = createLambert("RIG_RProxy_Material", [1, 0, 0])
		baseMat, baseSG = createLambert("RIG_BaseProxy_Material", [.5, .5, .5])
		
		buildLowPoly(bindCog, lowPolys, leftSG, centerSG, rightSG, baseSG)
		lowPolyGroup = pc.group(em = True, w = True, n = "ProxyGeoGroup")
		pc.parent(lowPolyGroup, doNotTouchGroup)
		lockAndHideTransform(lowPolyGroup)
		for proxy in lowPolys:
			pc.parent(proxy, lowPolyGroup)
		
		#throw 'em in a display layer
		pc.select(lowPolyGroup)
		layer = pc.createDisplayLayer(n = "ProxyGeo_DL", nr = True)
		layer.displayType.set(2)
		jointLayer.v.set(False) #if we have proxys, hide the joints
		
	#deselect everything
	pc.select(cl = True)
	
def buildLowPoly(root, lowPolys, leftSG, centerSG, rightSG, baseSG):
	"""
	build the proxy geometry recursively
	
	root is the top of the current joint hierarchy we are building
	lowPolys is an empty group that will have the geometry added to it
	*SG are shading group nodes for materials
	
	"""
	
	if root.side.get() == kSideLeft:
		curSG = leftSG
	elif root.side.get() == kSideRight:
		curSG = rightSG
	else:
		curSG = centerSG
	
	#build a sphere for the joint
	sphere = pc.polySphere(ax = [1,0,0], name = root.nodeName() + "JointGeo", r = .3)[0]
	pc.sets(curSG, e=True, forceElement=sphere)
	#assignColorOverride(sphere, root.side.get())
	constraint = pc.parentConstraint(root, sphere)
	constraint.attr("lockOutput").set(True)
	constraint = pc.scaleConstraint(root, sphere)
	constraint.attr("lockOutput").set(True)	
	
	lowPolys.append(sphere)
	
	rootPos = root.getTranslation("world")
	
	#loop over the children joints and build meshes for them too
	#also, build connective blocks between the root and each child
	for child in root.getChildren():
		if pc.nodeType(child) == "joint":
			childPos = child.getTranslation("world")
			
			#determine the length of the cube between the two joints
			axis = childPos - rootPos
			cubePos = rootPos + .5 * axis
			cubeHeight = pcd.length(axis)
			
			#make a cube between the joints
			cube = pc.polyCube(ax = axis, ch = True, name = root.nodeName() + "BoneGeo", h = cubeHeight, w = .3, d = .3)[0]
			cube.setTranslation(cubePos, "world")
			pc.sets(baseSG, e=True, forceElement=cube)
			constraint = pc.parentConstraint(root, cube, mo = True)
			constraint.attr("lockOutput").set(True)
			constraint = pc.scaleConstraint(root, cube, mo = True)
			constraint.attr("lockOutput").set(True)
			
			lowPolys.append(cube)			
			
			#build the low poly mesh for the child
			buildLowPoly(child, lowPolys, leftSG, centerSG, rightSG, baseSG)

def createLambert(name, color):
	"""create a lamber material with the given name and diffuse color"""
	mat = pc.shadingNode('lambert', asShader = True,  name = name)
	sg = pc.sets(renderable = True, noSurfaceShader = True, empty = True, name = name +"SG")
	pc.connectAttr(mat.outColor, sg.surfaceShader, force=True)
	mat.setColor(color)
	return mat, sg

def lockAndHideTransform(node, translate = True, rotate = True, scale = True, vis = True):
	"""Locks and hides the transform attributes of node based on the input options"""
	if node:
		if translate:
			pc.setAttr(node.tx, k = False, l = True, cb = False)
			pc.setAttr(node.ty, k = False, l = True, cb = False)
			pc.setAttr(node.tz, k = False, l = True, cb = False)
		
		if rotate:
			pc.setAttr(node.rx, k = False, l = True, cb = False)
			pc.setAttr(node.ry, k = False, l = True, cb = False)
			pc.setAttr(node.rz, k = False, l = True, cb = False)
		
		if scale:
			pc.setAttr(node.sx, k = False, l = True, cb = False)
			pc.setAttr(node.sy, k = False, l = True, cb = False)
			pc.setAttr(node.sz, k = False, l = True, cb = False)
		
		if vis:
			pc.setAttr(node.v, k = False, l = True, cb = False)
	
def connectBindSkeleton(driver, driven, bindConstraints):
	"""
	Recursively bind the driven joint to the driver joint
	Renames the driven joints to have a bind_ prefix followed by the driver name
	bindConstraints is a list to which any created constraints should be added
	
	"""
	
	driverChildren = driver.getChildren()
	drivenChildren = driven.getChildren()
	
	if len(driverChildren) != len(drivenChildren):
		pyError("Error in connectBindSkeleton() : non-equal number of children")
		return
	
	for i in range(len(driverChildren)):
		#loop through each pair of children and call recursively
		connectBindSkeleton(driverChildren[i], drivenChildren[i], bindConstraints)
	
	#finally, constrain the two passed in
	if driver and driven and pc.nodeType(driver) == "joint" and pc.nodeType(driven) == "joint":
		#rename
		pc.rename(driven, "bind_" + driver.nodeName())
		#pc.rename(driver, "ctrl_" + driver.nodeName()) #//*mayhaps put this back?
		
		parentConstraint = pc.parentConstraint(driver, driven, mo = True)
		parentConstraint.attr("lockOutput").set(True)
		parentConstraint.attr("interpType").set(2) #shortest interpolation type
		scaleConstraint = pc.scaleConstraint(driver, driven, mo = True)
		scaleConstraint.attr("lockOutput").set(True)
		
		bindConstraints.append(parentConstraint)
		bindConstraints.append(scaleConstraint)

def buildDefaultSkeleton(useEyes = True, useJaw = True, numNecks = 1, numSpines = 3,\
						xFrontAxis = False, armFkIk = 2, legFkIk = 1, createProxy = True):
	"""
	construct the basic biped skeleton template.
	At the end, the new cog joint is selected
	
	useEyes - Adds two eyes coming off the head joint
	useJaw - Adds a jaw coming off the head joint
	numNecks - The number of necks to create between the spine and head - should be more than 0
	numSpine - The number of spine joints to includes - should be more than 0
	xFrontAxis - if true, the skeleton is created facing down the x-axis, false down the z-axis
	armFkIk - should the arm be fk only (0), ik only (1), or blend between both(2)?
	legFkIk - should the leg be fk only (0), ik only (1), or blend between both(2)?
	createProxy - should proxy geometry be built?
	
	"""
	
	#create each of the basic parts
	cog = createCog(xFrontAxis, createProxy)
	hips = createHips()
	leftLeg = createLeg(kSideLeft, legFkIk)
	rightLeg = createLeg(kSideRight, legFkIk)
	spine, torsoRoot = createSpine(numSpines)
	neck, head = createHead(numNecks)
	leftClav = createClavicle(kSideLeft)
	leftArm = createArm(kSideLeft, armFkIk)
	rightClav = createClavicle(kSideRight)
	rightArm = createArm(kSideRight, armFkIk)
	
	#create the optional parts
	if useEyes:
		leftEye = createEye(kSideLeft)
		rightEye = createEye(kSideRight)
	if useJaw:
		jaw = createJaw()
	
	
	#parent everything together into the required hierarchy
	pc.parent([hips, spine], cog)
	pc.parent([leftLeg, rightLeg], hips)
	pc.parent([neck, leftClav, rightClav], torsoRoot)
	pc.parent(leftArm, leftClav)
	pc.parent(rightArm, rightClav)
	
	if useEyes:
		pc.parent([leftEye, rightEye], head)
	if useJaw:
		pc.parent(jaw, head)
	
	#move all the joints to be in vaguely a useful configuration
	hips.setTranslation([0, -1, 0], "object")
	leftLeg.setTranslation([2, -2, 0], "object")
	rightLeg.setTranslation([-2, -2, 0], "object")
	spine.setTranslation([0, 1, 0], "object")
	neck.setTranslation([0, 1, 0], "object")
	leftClav.setTranslation([2, 0, 0], "object")
	leftArm.setTranslation([2, 0, 0], "object")
	rightClav.setTranslation([-2, 0, 0], "object")
	rightArm.setTranslation([-2, 0, 0], "object")
	
	if useEyes:
		leftEye.setTranslation([1, 1, 1.5], "object")
		rightEye.setTranslation([-1, 1, 1.5], "object")
	if useJaw:
		jaw.setTranslation([0, -1, 1], "object")
	
	cog.setTranslation([0, 9, 0], "object")
	
	#rotate the whole thing if needed
	if xFrontAxis:
		cog.ry.set(90)
		pc.makeIdentity(cog, apply = True, r = True)
	
	#select the cog
	pc.select(cog)

def buildSingleJoint(inJoint, buildList):
	"""
	Builds a single-joint component, with nothing special to it
	
	Also builds any child compoenents. Deletes inJoint at the end and returns the
	newly built joint
	
	"""
	
	#get basic info
	jointPos = inJoint.getTranslation("world")
	jointName = str(inJoint)
	jointSide = inJoint.side.get()
	jointType = inJoint.otherType.get()
	
	toParent = []
	
	#build any child components
	for child in inJoint.getChildren():
		returned = buildComponent(child, buildList)
		toParent.append(returned)
	
	pc.delete(inJoint)
	
	#make a new joint at the location of the old one
	pc.select(cl = True)
	outJoint = pc.joint(p = jointPos, absolute = True, name = jointName)
	labelJoint(outJoint, jointType, jointSide, False)
	
	#parent the built children under the new joint
	for child in toParent:
		if child:
			pc.parent(child, outJoint)
	
	return outJoint

def buildSpine(inSpine, buildList):
	"""
	Builds a spine component.
	
	Also builds any child compoenents. Deletes inSpine at the end and returns the
	newly built spine from the root
	
	"""
	
	#since we can have an unknown number, build the spine recursively
	outSpine =  buildSpineRecursive(inSpine, buildList)
	
	buildList.append(SpineContainer(outSpine))
	return outSpine

def buildSpineRecursive(inSpine, buildList):
	"""
	Builds the spine from inSpine.
	Recursively calls this function on any spine children.
	
	"""
	
	pc.select(cl = True)
	
	#make sure the input is right
	if not (inSpine and pc.nodeType(inSpine) == "joint"):
			return None
			
	jointType = inSpine.otherType.get()
		
	#if it's not a spine or torso_root, we're done
	if not (jointType == kSpineName or jointType == kTorsoRootName):
		return None
	
	toParent = []
	#otherwise, forge ahead
	for child in inSpine.getChildren():
		#if we find a spine or torso_root, save it, otherwise build
		if pc.nodeType(child) == "joint" and (child.otherType.get() == kSpineName or child.otherType.get() == kTorsoRootName):
			toParent.append(buildSpineRecursive(child, buildList))
		else:
			toParent.append(buildComponent(child, buildList))
	
	#now we can make the new version of the joint
	#first grab it's info
	jointPos = inSpine.getTranslation("world")
	jointName = inSpine.nodeName()
	jointSide = inSpine.side.get()
	sidePrefix = getSidePrefix(jointSide)
	jointType = inSpine.otherType.get()
	
	pc.delete(inSpine)
	
	pc.select(cl = True)
	outJoint = pc.joint(p = jointPos, absolute = True, name = sidePrefix+jointName)
	labelJoint(outJoint, jointType, jointSide, False)
	
	for child in toParent:
		if child:
			pc.parent(child, outJoint)
	
	return outJoint


def buildHead(inNeck, buildList):
	"""
	Builds a head component, along with the preceeding neck joints.
	
	Also builds any child compoenents. Deletes inNeck at the end and returns the
	newly built neck from the root
	
	"""
	
	#arbitray number of necks, so do it recursively
	outNeck =  buildHeadRecursive(inNeck, buildList)
	
	buildList.append(HeadContainer(outNeck))
	return outNeck

def buildHeadRecursive(inHead, buildList):
	"""
	Builds the head/neck from inHead.
	Recursively calls this function on any neck/head children.
	
	"""	
	
	pc.select(cl = True)
	
	#make sure the input is right
	if not (inHead and pc.nodeType(inHead) == "joint"):
			return None
			
	jointType = inHead.otherType.get()
		
	#if it's not a neck, head, or head_tip, we're done
	if not (jointType == kNeckName or jointType == kHeadName or jointType == kHeadTipName):
		return None
	
	toParent = []
	#otherwise, forge ahead
	for child in inHead.getChildren():
		#if we find a neck, head, or head_tip, save it, otherwise build
		if pc.nodeType(child) == "joint":
			childJointType = child.otherType.get()
			if (childJointType == kNeckName or childJointType == kHeadName or childJointType == kHeadTipName):
				toParent.append(buildHeadRecursive(child, buildList))
			else:
				toParent.append(buildComponent(child, buildList))
		else:
			toParent.append(buildComponent(child, buildList))
	
	#now we can make the new version of the joint
	#first grab it's info
	jointPos = inHead.getTranslation("world")
	jointName = inHead.nodeName()
	jointSide = inHead.side.get()
	jointType = inHead.otherType.get()
	
	pc.delete(inHead)
	
	pc.select(cl = True)
	outJoint = pc.joint(p = jointPos, absolute = True, name = jointName)
	labelJoint(outJoint, jointType, jointSide, False)
	
	for child in toParent:
		if child:
			pc.parent(child, outJoint)
	
	return outJoint
		

def buildCog(inCog, buildList):
	"""
	Builds a cog component, and sets attributes on the buildList
	based on attributes on the inCog. Deletes the inCog on completion
	
	"""
	
	buildList.setFrontX(inCog.attr(kCogAlignXAttr).get())
	buildList.setCreateProxy(inCog.attr(kCogCreateProxyAttr).get())
	outCog = buildSingleJoint(inCog, buildList)
	
	pc.rename(outCog, kCogName)
	
	return outCog

def buildClavicle(inClav, buildList):
	"""
	Builds a clavicle component, naming it based on inClav and the clavicle's side.
	Deletes inClav. Builds any children of inClav.	
	
	"""	
	
	side = inClav.side.get()
	sidePrefix = getSidePrefix(side)
	outClav = buildSingleJoint(inClav, buildList)
	
	pc.rename(outClav, sidePrefix + kClavicleName)
	
	buildList.append(ClavicleContainer(outClav))
	return outClav

def buildJaw(inJaw, buildList):
	"""
	Builds a jaw component.
	Deletes inJaw. Builds any children of inJaw or the jaw tip joint.	
	
	"""	
	
	#first find the tip
	inTip = None
	for child in inJaw.getChildren():
		if pc.nodeType(child) == "joint" and child.otherType.get() == kJawTipName:
			inTip = child
			break
	
	outTip = None
	if inTip:
		outTip = buildSingleJoint(inTip, buildList)
	
	outJaw = buildSingleJoint(inJaw, buildList)
	
	if outTip:
		pc.parent(outTip, outJaw)
	
	buildList.append(JawContainer(outJaw))
	
	return outJaw

def buildEye(inEye, buildList):
	"""
	Builds an eye component.
	Deletes inEye. Builds any children of inEye.	
	
	"""	
	
	#if it has been visited, delete it and return it's connection
	if inEye.hasAttr(kEyeVisitedAttr):
		inputs = inEye.attr(kEyeVisitedAttr).inputs()
		pc.delete(inEye)
		if inputs:
			return inputs[0]
		else:
			return None
	
	theContainer = EyeContainer()
	
	#first handle this eye
	outEye = handleEye(inEye, theContainer, buildList)
	
	#now, go through all children of our parent and grab all eyes
	parent = inEye.getParent()
	
	if parent:
		for child in parent.getChildren():
			if pc.nodeType(child) == "joint" and child.otherType.get() == kEyeName and child != inEye:
				handleEye(child, theContainer, buildList)
	
	buildList.append(theContainer)
	
	pc.delete(inEye)
	
	return outEye

def handleEye(inJoint, container, buildList):
	"""
	Handles the actual building of the eye. Since all eyes under a given
	joint must be grouped together, they are added to the container here.
	
	"""
	
	jointPos = inJoint.getTranslation("world")
	jointName = str(inJoint)
	jointSide = inJoint.side.get()
	jointType = inJoint.otherType.get()
	
	toParent = []
	
	for child in inJoint.getChildren():
		returned = buildComponent(child, buildList)
		toParent.append(returned)
	
	#pc.delete(inJoint) #done in buildEye as the eye is actually found
	
	pc.select(cl = True)
	outJoint = pc.joint(p = jointPos, absolute = True, name = jointName)
	labelJoint(outJoint, jointType, jointSide, False)
	
	for child in toParent:
		if child:
			pc.parent(child, outJoint)
			
	#set the eye as visited and connect to the output
	inJoint.addAttr(kEyeVisitedAttr, at = "message")
	pc.connectAttr(outJoint.message, inJoint.attr(kEyeVisitedAttr), f= True)
	
	container.addEye(outJoint)
	
	return outJoint

def buildHips(inHips, buildList):
	"""
	Builds an hips component.
	Deletes inHips. Builds any children of inHips.
	
	The position of the new hip joint is halfway between inHips and the
	parent of inHips. A new temporary joint is created as a child of
	the output hips at the location of inHips to help with orientation later.
	
	"""	
	
	jointPos = inHips.getTranslation("world")
	jointName = str(inHips)
	jointSide = inHips.side.get()
	jointType = inHips.otherType.get()
	
	#we need the hips' parent's location for our little orientation trick
	parentPos = inHips.getParent().getTranslation("world")
	parentVect = jointPos - parentPos
	midPos = parentPos + 0.5 * parentVect #we use this for the hips for now
	
	toParent = []
	
	#build children
	for child in inHips.getChildren():
		returned = buildComponent(child, buildList)
		toParent.append(returned)
	
	pc.delete(inHips)
	
	#make the two joints
	pc.select(cl = True)
	outHips = pc.joint(p = midPos, absolute = True, name = jointName)
	labelJoint(outHips, jointType, jointSide, False)
	
	pc.select(cl = True)
	tip = pc.joint(p = jointPos, absolute = True, name = kHipsTipName)
	labelJoint(tip, kHipsTipName, jointSide, False)
	
	pc.parent(tip, outHips)
	
	#parent any built children to the tip joint
	for child in toParent:
		if child:
			pc.parent(child, tip)
	
	buildList.append(HipsContainer(outHips))
	
	pc.rename(outHips, kHipsName)
	
	return outHips
	

def buildArm(inShoulder, buildList = None):
	"""
	Builds an arm component rooted at inShoulder.
	Also builds any children of arm joints.
	Inserts a roll joint between the elbow and wrist.
	Deletes the inShoulder hierarchy.
	
	
	"""
	
	#build a container to find all the arm joints and build any children
	inArm = ArmContainer(inShoulder, buildList)
	if not inArm.hasAllJoints:
		pyError("buildArm() :: Not all joints found")
		return		
	
	#if we get here, we found all limbs so find the world space translations
	shoulderPos = inArm.shoulder.getTranslation("world")
	elbowPos = inArm.elbow.getTranslation("world")
	wristPos = inArm.wrist.getTranslation("world")
	handPos = inArm.hand.getTranslation("world")
	
	#get extra info
	side = inArm.shoulder.side.get()
	sidePrefix = getSidePrefix(side)
	fkIk = inShoulder.attr(kArmFkIkAttr).get()
	
	#calculate the forearm position
	forearmPos = (wristPos - elbowPos)/2.0 + elbowPos
	
	pc.delete(inShoulder)
	
	#make the new joints
	pc.select(cl=True)
	shoulder = pc.joint(p = shoulderPos, absolute = True, name = sidePrefix + kShoulderName)
	labelJoint(shoulder, kShoulderName, side, False)
	elbow = pc.joint(p = elbowPos, absolute = True, name = sidePrefix + kElbowName)
	labelJoint(elbow, kElbowName, side, False)
	forearm = pc.joint(p = forearmPos, absolute = True, name = sidePrefix + kForearmName)
	labelJoint(forearm, kForearmName, side, False)
	wrist = pc.joint(p = wristPos, absolute = True, name = sidePrefix + kWristName)
	labelJoint(wrist, kWristName, side, False)
	hand = pc.joint(p = handPos, absolute = True, name = sidePrefix + kHandName)
	labelJoint(hand, kHandName, side, False)
	
	
	#parent anything that was built to them
	for child in inArm.shoulderList:
		pc.parent(child, shoulder)
	for child in inArm.elbowList:
		pc.parent(child, elbow)
	for child in inArm.forearmList:
		pc.parent(child, forearm)
	for child in inArm.wristList:
		pc.parent(child, wrist)
	for child in inArm.handList:
		pc.parent(child, hand)
		
	#add the arm to the buildList
	buildList.append(ArmContainer(shoulder, fkIk = fkIk))
	
	return shoulder

def buildFkArm(inArm):
	"""
	Sets up an fk arm control rig for the arm in the container inArm.
	Returns a new container for the fk rig.
	Duplicates the inArm from the shoulder and constrains inArm to the duplicate.
	
	Each joint (shoulder, elbow, wrist) gets a circle controller. The wrist circle
	controller is setup to use the forearm joint as a roll joint.
	
	"""
	
	outArm = inArm.duplicate() #make a duplicate	
	inArm.constrainTo(outArm) #constrain the inArm to the new arm
	
	#create controllers for the new arm
	shoulderCTRL, shoulderGroup = attachController(outArm.shoulder, name = inArm.shoulder.nodeName() + "CTRL")
	assignColorOverride(shoulderCTRL, outArm.shoulder.side.get())
	elbowCTRL, elbowGroup = attachController(outArm.elbow, name = inArm.elbow.nodeName() + "CTRL")
	assignColorOverride(elbowCTRL, outArm.shoulder.side.get())
	#pc.setAttr(elbowCTRL.rx, k = False, l = True, cb = False)
	#pc.setAttr(elbowCTRL.rz, k = False, l = True, cb = False)
	
	#special for the wrist/forearm
	controller, controllerGroup = attachController(outArm.wrist,\
												toSecondParent = True, skip = "x", name = inArm.wrist.nodeName() + "CTRL")
	pc.orientConstraint(controller, outArm.forearm, mo = True, skip = ["y", "z"])
	assignColorOverride(controller, outArm.shoulder.side.get())
	
	#hide the joints
	pc.hide([outArm.shoulder])
	
	#store the controllers
	outArm.shoulderCtrl = shoulderCTRL
	outArm.elbowCtrl = elbowCTRL
	outArm.handCtrl = controller
	
	#add an attribute to all the new joints signifying them as fk, and rename them
	for joint in outArm.getJointList():
		joint.addAttr(kFkJointAttr, at = "bool")
		joint.rename(joint.nodeName() + kFkPrefix)
	
	#fix the name for the shoulder
	outArm.shoulder.rename(inArm.shoulder.nodeName() + kFkPrefix)
	
	outArm.addGroups(shoulderGroup)
	outArm.addGroups(elbowGroup)
	outArm.addGroups(controllerGroup)
	
	return outArm
	
def buildIkArm(inArm):
	"""
	Sets up an ik arm control rig for the arm in the container inArm.
	Returns a new container for the fk rig.
	Duplicates the inArm from the shoulder and constrains inArm to the duplicate.
	
	An rp ik chain is created from the shoulder to the forearm, with the effector at the wrist.
	An sc ik chain goes from the wrist to the hand to maintain the hand's orientation.
	A box controller is placed at the wrist to control the ik. The forearm is setup as a roll joint.
	Elbow position is controlled from a twist attribute on the controller instead of a pole-vector.
	Tha hand can be optionally pinned to the shoulder's parent through a constraint.
	
	"""	
	
	outArm = inArm.duplicate() #make a duplicate
	inArm.constrainTo(outArm) #constrain the inArm to the new arm
	
	side = inArm.shoulder.side.get()
	sidePrefix = getSidePrefix(side)
	
	#create IK handle
	wristHandle = pc.ikHandle(startJoint = outArm.shoulder, ee = outArm.forearm, solver = "ikRPsolver")[0]
	
	#get the end effector and adjust the scale and rotate pivots to the wrist
	effector = wristHandle.getEndEffector()
	effector.setRotatePivot(outArm.wrist.getTranslation("world"), "world")
	#setScalePivot is bugged in Pymel and can't be called, so we just copy the rotate pivot manually
	effector.scalePivot.set(effector.rotatePivot.get())
	#effector.setScalePivot(outArm.wrist.getTranslation("world"), "world")
	
	#create the hand IK handle
	handHandle = pc.ikHandle(startJoint = outArm.wrist, ee = outArm.hand, solver = "ikSCsolver")[0]
	
	#make the hand controller
	CTRL, CTRLGroup = attachController(outArm.wrist, kBoxCurve, toParent = False, drive = False, hideTranslate = False,\
									name = sidePrefix + "ArmCTRL")
	
	assignColorOverride(CTRL, side)
	
	#now, set up the two orient constraints (forearm and wrist)
	constraint = pc.orientConstraint(CTRL, outArm.forearm, mo = True, skip = ["y", "z"])
	constraint.attr("lockOutput").set(True)
	
	#parent constrain the IK handles to the controller
	parentConstraint = pc.parentConstraint(CTRL, wristHandle, mo = True)
	parentConstraint.attr("lockOutput").set(True)
	parentConstraint = pc.parentConstraint(CTRL, handHandle, mo = True)
	parentConstraint.attr("lockOutput").set(True)
	
	#create a twist attribute for the controller
	CTRL.addAttr("elbowTwist", k = True)
	pc.connectAttr(CTRL.elbowTwist, wristHandle.twist, force = True)
	
	#if the shoulder has a parent, constrain the CTRLGroup to it
	shoulderParent = outArm.shoulder.getParent()
	if(shoulderParent):
		parentConstraint = pc.parentConstraint(shoulderParent, CTRLGroup, mo = True)
		parentConstraint.attr("lockOutput").set(True)
		
		CTRL.addAttr("pin", k = True, dv = 1, min = 0, max = 1, at = "long")
		targets = parentConstraint.getTargetList()
		weights = parentConstraint.getWeightAliasList()
		numTargets = len(targets)
		
		for i in range(numTargets):
			if targets[i] == shoulderParent:
				pc.setDrivenKeyframe(weights[i], cd = CTRL.pin, dv = 0, v = 1, ott = "linear")
				pc.setDrivenKeyframe(weights[i], cd = CTRL.pin, dv = 1, v = 0, ott = "linear")
	
	#hide the handles and joints
	pc.hide([wristHandle, handHandle, outArm.shoulder])
	
	#store the controller
	outArm.handCtrl = CTRL
	
	#add an attribute to all the new joints signifying them as ik
	for joint in outArm.getJointList():
		joint.addAttr(kIkJointAttr, at = "bool")
		joint.rename(joint.nodeName() + kIkPrefix)
	
	#fix the name for the shoulder
	outArm.shoulder.rename(inArm.shoulder.nodeName() + kIkPrefix)
	
	outArm.addGroups(CTRLGroup)
	outArm.addGroups(wristHandle)
	outArm.addGroups(handHandle)
	
	pc.rename(wristHandle, sidePrefix + "WristHandle")
	pc.rename(handHandle, sidePrefix + "HandHandle")
	
	return outArm
	
	
	
def buildLeg(inFemur, buildList = None):
	"""
	Builds a leg component rooted at inFemur.
	Also builds any children of leg joints.
	Deletes the inShoulder hierarchy.
	
	
	"""	
	
	inLeg = LegContainer(inFemur, buildList)
	if not inLeg.hasAllJoints:
		pyError("buildLeg() :: Not all joints found")
		return	
		
	
	#if we get here, we found all limbs so find the world space translations
	femurPos = inLeg.femur.getTranslation("world")
	kneePos = inLeg.knee.getTranslation("world")
	anklePos = inLeg.ankle.getTranslation("world")
	ballPos = inLeg.ball.getTranslation("world")
	toePos = inLeg.toe.getTranslation("world")
	
	#get extra info
	side = inLeg.femur.side.get()
	sidePrefix = getSidePrefix(side)
	fkIk = inFemur.attr(kLegFkIkAttr).get()
	
	#find the heel joint and store it's position
	heelPos = None
	for child in inLeg.ankle.getChildren():
		if pc.nodeType(child) == "joint" and child.otherType.get() == kHeelName:
			heelPos = child.getTranslation("world")
			break
	
	if not heelPos:
		pyError("Heel Position missing!")
		return
	
	pc.delete(inFemur)
	
	#make the new joints
	pc.select(cl=True)
	femur = pc.joint(p = femurPos, absolute = True, name = sidePrefix + kFemurName)
	labelJoint(femur, kFemurName, side, False)
	knee = pc.joint(p = kneePos, absolute = True, name = sidePrefix + kKneeName)
	labelJoint(knee, kKneeName, side, False)
	ankle = pc.joint(p = anklePos, absolute = True, name = sidePrefix + kAnkleName)
	labelJoint(ankle, kAnkleName, side, False)
	ball = pc.joint(p = ballPos, absolute = True, name = sidePrefix + kBallName)
	labelJoint(ball, kBallName, side, False)
	toe = pc.joint(p = toePos, absolute = True, name = sidePrefix + kToeName)
	labelJoint(toe, kToeName, side, False)
	
	femur.orientJoint("xyz", sao = "yup", ch = True, zso = True)
	
	baseLeg = LegContainer(femur, fkIk = fkIk)
	baseLeg.heelPos = heelPos
	
	#parent anything that was built to them
	for child in inLeg.femurList:
		pc.parent(child, femur)
	for child in inLeg.kneeList:
		pc.parent(child, knee)
	for child in inLeg.ankleList:
		pc.parent(child, ankle)
	for child in inLeg.ballList:
		pc.parent(child, ball)
	for child in inLeg.toeList:
		pc.parent(child, toe)
		
	#add the arm to the buildList
	buildList.append(baseLeg)
	
	return femur

def buildFkLeg(inLeg):
	"""
	Sets up an fk leg control rig for the leg in the container inLeg.
	Returns a new container for the fk rig.
	Duplicates the inLeg from the femur and constrains inLeg to the duplicate.
	
	Each joint (femur, knee, ankle) gets a circle controller.
	
	"""
	
	outLeg = inLeg.duplicate() #make a duplicate
	inLeg.constrainTo(outLeg) #constrain the inLeg to the new leg

	side = outLeg.femur.side.get()

	#create controllers for the new leg
	femurCTRL, femurGroup = attachController(outLeg.femur, name = inLeg.femur.nodeName() + "CTRL")
	assignColorOverride(femurCTRL, side)
	kneeCTRL, kneeGroup = attachController(outLeg.knee, name = inLeg.knee.nodeName() + "CTRL")
	assignColorOverride(kneeCTRL, side)
	#kneeCTRL.rx.lock()
	#kneeCTRL.rz.lock()
	
	ankleCTRL, ankleGroup = attachController(outLeg.ankle, name = inLeg.ankle.nodeName() + "CTRL")
	assignColorOverride(ankleCTRL, side)
	ballCTRL, ballGroup = attachController(outLeg.ball, name = inLeg.ball.nodeName() + "CTRL")
	assignColorOverride(ballCTRL, side)
	
	#hide the joints
	pc.hide([outLeg.femur])
	
	outLeg.femurCtrl = femurCTRL
	outLeg.kneeCtrl = kneeCTRL
	outLeg.ankleCtrl = ankleCTRL
	outLeg.ballCtrl = ballCTRL
	
	#add an attribute to all the new joints signifying them as ik
	for joint in outLeg.getJointList():
		joint.addAttr(kFkJointAttr, at = "bool")
		joint.rename(joint.nodeName() + kFkPrefix)
	
	#fix the name for the shoulder
	outLeg.femur.rename(inLeg.femur.nodeName() + kFkPrefix)
	
	outLeg.addGroups(femurGroup)
	outLeg.addGroups(kneeGroup)
	outLeg.addGroups(ankleGroup)
	outLeg.addGroups(ballGroup)
	
	return outLeg

def buildIkLeg(inLeg, heelPos):
	"""
	Sets up an ik leg control rig for the leg in the container inLeg.
	Returns a new container for the fk rig.
	Duplicates the inLeg from the shoulder and constrains inLeg to the duplicate.
	
	An rp ik chain is created from the femur to the ankle.
	An sc ik chain goes from the ankle to the ball, and another from the ball to the toe.
	A box controller is placed at the ankle to control the ik.
	Knee position is controlled from a twist attribute on the controller instead of a pole-vector.
	A group-based reverse foot rig is created.
	The leg can be optionally pinned to the femur's parent through a constraint.
	
	heelPos - a vector coresponding to the heel pivot position for the reverse foor rig.
	
	"""	
	
	outLeg = inLeg.duplicate() #make a duplicate
	inLeg.constrainTo(outLeg) #constrain the inLeg to the new leg
	
	side = inLeg.femur.side.get()
	sidePrefix = getSidePrefix(side)
	
	#create IK handles
	ankleHandle = pc.ikHandle(startJoint = outLeg.femur, ee = outLeg.ankle, solver = "ikRPsolver")[0]
	ballHandle = pc.ikHandle(startJoint = outLeg.ankle, ee = outLeg.ball, solver = "ikSCsolver")[0]
	toeHandle = pc.ikHandle(startJoint = outLeg.ball, ee = outLeg.toe, solver = "ikSCsolver")[0]
	
	#make the hand controller
	CTRL, CTRLGroup = attachController(outLeg.ball, kBoxCurve, toParent = False, drive = False, hideTranslate = False, \
									name = sidePrefix + "LegCTRL")
	
	assignColorOverride(CTRL, side)
	
	#make groups for the reverse foot
	heelPivot = pc.group(n = "Heel_Pivot", em = True)
	heelPivot.setTranslation(heelPos, "world")
	
	
	toePivot = pc.group(n = "Toe_Pivot", em = True)
	toePivot.setTranslation(outLeg.toe.getTranslation("world"), "world")
	
	
	ankleGroup = pc.group(n = "Ankle_Group", em = True)
	ankleGroup.setTranslation(outLeg.ball.getTranslation("world"), "world")
	
	
	toeGroup = pc.group(n = "Toe_Group", em = True)
	toeGroup.setTranslation(outLeg.ball.getTranslation("world"), "world")
	
	
	#now make the hierarchy
	pc.parent(toeGroup, toePivot)
	pc.parent(ankleGroup, toePivot)
	pc.parent(toePivot, heelPivot)
	pc.parent(heelPivot, CTRL)
	
	pc.makeIdentity(toeGroup, apply = True, r = True, t = True) #freeze the transforms
	pc.makeIdentity(ankleGroup, apply = True, r = True, t = True) #freeze the transforms
	pc.makeIdentity(toePivot, apply = True, r = True, t = True) #freeze the transforms
	pc.makeIdentity(heelPivot, apply = True, r = True, t = True) #freeze the transforms
	
	pc.parent(toeHandle, toeGroup)
	pc.parent(ankleHandle, ankleGroup)
	pc.parent(ballHandle, toePivot)
	
	#create a attributes for the reverse foot
	CTRL.addAttr("heelPivot", k = True, min = -90, max = 90)
	pc.connectAttr(CTRL.heelPivot, heelPivot.rz, force = True)
	CTRL.addAttr("heelTwist", k = True, min = -90, max = 90)
	pc.connectAttr(CTRL.heelTwist, heelPivot.ry, force = True)
	CTRL.addAttr("heelLift", k = True, min = -90, max = 0)
	pc.connectAttr(CTRL.heelLift, ankleGroup.rz, force = True)
	CTRL.addAttr("toePivot", k = True, min = -90, max = 0)
	pc.connectAttr(CTRL.toePivot, toePivot.rz, force = True)
	CTRL.addAttr("toeTwist", k = True, min = -90, max = 90)
	pc.connectAttr(CTRL.toeTwist, toePivot.ry, force = True)
	CTRL.addAttr("toeWiggle", k = True, min = -90, max = 90)
	pc.connectAttr(CTRL.toeWiggle, toeGroup.rz, force = True)
	
	#create a twist attribute for the knee
	CTRL.addAttr("kneeTwist", k = True)
	pc.connectAttr(CTRL.kneeTwist, ankleHandle.twist, force = True)
	
	#if the femur has a parent, constrain the CTRLGroup to it
	femurParent = outLeg.femur.getParent()
	if(femurParent):
		parentConstraint = pc.parentConstraint(femurParent, CTRLGroup, mo = True)
		parentConstraint.attr("lockOutput").set(True)
		
		CTRL.addAttr("pin", k = True, dv = 1, min = 0, max = 1, at = "long")
		targets = parentConstraint.getTargetList()
		weights = parentConstraint.getWeightAliasList()
		numTargets = len(targets)
		
		for i in range(numTargets):
			if targets[i] == femurParent:
				pc.setDrivenKeyframe(weights[i], cd = CTRL.pin, dv = 0, v = 1, ott = "linear")
				pc.setDrivenKeyframe(weights[i], cd = CTRL.pin, dv = 1, v = 0, ott = "linear")
	
	#hide the handles and joints
	pc.hide([ankleHandle, ballHandle, toeHandle, outLeg.femur])
	
	pc.rename(toeHandle, sidePrefix + "ToeHandle")
	pc.rename(ankleHandle, sidePrefix + "BallHandle")
	pc.rename(ballHandle, sidePrefix + "AnkleHandle")
	
	#store the controller
	outLeg.footCtrl = CTRL
	
	#add an attribute to all the new joints signifying them as ik
	for joint in outLeg.getJointList():
		joint.addAttr(kIkJointAttr, at = "bool")
		joint.rename(joint.nodeName() + kIkPrefix)
	
	#fix the name for the shoulder
	outLeg.femur.rename(inLeg.femur.nodeName() + kIkPrefix)
	
	outLeg.addGroups(CTRLGroup)
	
	return outLeg

def createArm(side = kSideCenter, fkIk = kFkIk):
	"""
	Create an arm component labeled with the given side
	and set to the given ik/fk choice.
	
	Returns the arm's root joint.
	
	"""
	
	multiplier = 1
	
	#adjust for right side arms
	if side == kSideRight:
		multiplier = -1
	
	pc.select(cl=True)
	shoulder = pc.joint(p = [0, 0, 0], relative = True, name = kShoulderName)
	labelJoint(shoulder, kShoulderName, side)
	
	shoulder.addAttr(kArmFkIkAttr, at = "short", dv = fkIk, minValue = 0, maxValue = 2)
	
	#position and label the joints
	elbow = pc.joint(p = [multiplier * 2, 0, -.1], relative = True, name = kElbowName)
	labelJoint(elbow, kElbowName, side)
	wrist = pc.joint(p = [multiplier * 2, 0, .1], relative = True, name = kWristName)
	labelJoint(wrist, kWristName, side)
	hand = pc.joint(p = [multiplier * 1, 0, 0], relative = True, name = kHandName)
	labelJoint(hand, kHandName, side)
	pc.select(shoulder)
	return shoulder

def createLeg(side = kSideCenter, fkIk = kIkOnly):
	"""
	Create a leg component labeled with the given side
	and set to the given ik/fk choice.
	
	Returns the leg's root joint.
	
	"""
	
	pc.select(cl=True)
	femur = pc.joint(p = [0, 0, 0], relative = True, name = kFemurName)
	labelJoint(femur, kFemurName, side)
	
	femur.addAttr(kLegFkIkAttr, at = "short", dv = fkIk, minValue = 0, maxValue = 2)
	
	#position and label the joints
	knee = pc.joint(p = [0, -2, 1], relative = True, name = kKneeName)
	labelJoint(knee, kKneeName, side)
	ankle = pc.joint(p = [0, -2, -1], relative = True, name = kAnkleName)
	labelJoint(ankle, kAnkleName, side)
	
	heel = pc.joint(p = [0, -1, -1], relative = True, name = kHeelName)
	labelJoint(heel, kHeelName, side)
	
	pc.select(ankle)
	
	ball = pc.joint(p = [0, -1, 2], relative = True, name = kBallName)
	labelJoint(ball, kBallName, side)
	toe = pc.joint(p = [0, 0, 1], relative = True, name = kToeName)
	labelJoint(toe, kToeName, side)
	pc.select(femur)
	return femur

def createSpine(numJoints = 3, side = kSideCenter, name = ''):
	"""
	Create a spine component labeled with the given side.
	If name is empty, "spine" is used for the joint names.
	numJoints must be at least 1, and controls the number of spine joints
	in the chain.
	
	Returns the spine's root joint.
	
	"""
	
	pc.select(cl=True)
	
	if name == '' or None:
		name = kSpineName
		torsoName = kTorsoRootName
	else:
		torsoName = name + "_Tip"
	
	#make the first joint
	spineRoot = pc.joint(p = [0, 0, 0], relative = True, name = name + "1")
	labelJoint(spineRoot, kSpineName, side)
	
	for i in range(numJoints - 1):	
		joint = pc.joint(p = [0, 1, 0], relative = True, name = name + str(i + 2))
		labelJoint(joint, kSpineName, side)

	torsoRoot = pc.joint(p = [0, 1, 0], relative = True, name = torsoName)
	labelJoint(torsoRoot, kTorsoRootName, side)
	
	pc.select(spineRoot)
	return spineRoot, torsoRoot

def createHead(numJoints = 1):
	"""
	Create a head component labeled with the given side.
	numJoints must be at least 1, and controls the number of neck joints
	in the chain before the head joint.
	
	Returns the head's root neck joint.
	
	"""
	
	pc.select(cl=True)
	
	#make the first joint
	neckRoot = pc.joint(p = [0, 0, 0], relative = True, name = kNeckName + "1")
	labelJoint(neckRoot, kNeckName, kSideCenter)
	
	for i in range(numJoints - 1):	
		joint = pc.joint(p = [0, 1, 0], relative = True, name = kNeckName + str(i + 2))
		labelJoint(joint, kNeckName, kSideCenter)

	head = pc.joint(p = [0, 2, 0], relative = True, name = kHeadName)
	labelJoint(head, kHeadName, kSideCenter)
	
	headTip = pc.joint(p = [0, 2, 0], relative = True, name = kHeadTipName)
	labelJoint(headTip, kHeadTipName, kSideCenter)
	
	pc.select(neckRoot)
	return neckRoot, head

def createHips():
	"""
	Create a hips component and return it.
	
	"""
	
	pc.select(cl=True)
	
	#make the first joint
	hips = pc.joint(p = [0, 0, 0], relative = True, name = kHipsName)
	labelJoint(hips, kHipsName, kSideCenter)
	
	pc.select(hips)
	return hips

def createCog(alignX = False, createProxy = True):
	"""
	Create a cog component and return it.'
	
	alignX determines if the rig is aligned down the x-axis (true)
	or the z-axis (false).
	createProxy determines if proxy geometry is created when the rig is built.
	
	"""
	
	pc.select(cl=True)
	
	#make the first joint
	cog = pc.joint(p = [0, 0, 0], relative = True, name = kCogName)
	labelJoint(cog, kCogName, kSideCenter)
	
	cog.addAttr(kCogAlignXAttr, k = False, at = "bool", dv = alignX)
	cog.addAttr(kCogCreateProxyAttr, k = False, at = "bool", dv = createProxy)
	
	pc.select(cog)
	return cog

def createClavicle(side = kSideCenter):
	"""
	Create a hips component labeled with the given side and return it.
	
	"""
	
	pc.select(cl=True)
	
	#make the first joint
	clav = pc.joint(p = [0, 0, 0], relative = True, name = kClavicleName)
	labelJoint(clav, kClavicleName, side)
	
	pc.select(clav)
	return clav

def createJaw():
	"""
	Create a jaw component and return it.
	
	"""
	
	pc.select(cl=True)
	
	#make the first joint
	jaw = pc.joint(p = [0, 0, 0], relative = True, name = kJawName)
	labelJoint(jaw, kJawName, kSideCenter)
	
	tip = pc.joint(p = [0, 0, 1], relative = True, name = kJawTipName)
	labelJoint(tip, kJawTipName, kSideCenter)
	
	pc.select(jaw)
	return jaw

def createEye(side = kSideCenter):
	"""
	Create an eye component labelled with the given side and return it.
	
	"""
	
	pc.select(cl=True)
	
	#make the first joint
	eye = pc.joint(p = [0, 0, 0], relative = True, name = kEyeName)
	labelJoint(eye, kEyeName, side)
	
	pc.select(eye)
	return eye

def labelJoint(joint, label, side, draw = True):
	"""
	sets the joint label attributes on joint to be in "other"
	mode and assigns label to the "otherType" attribute.
	Set's the side attribute to side.
	Set's the drawLabel attribute to draw.
	
	"""
	
	joint.attr("type").set(18) #set it to "other" mode
	joint.otherType.set(label)
	joint.drawLabel.set(draw)
	joint.side.set(side)

#constrains the given object to a controller.
#If toParent is true, constrains the controller to the object's parent
def attachController(theObject, controlType = kCircleCurve, toParent = True, toSecondParent = False, skip = [], drive = True,\
					hideTranslate = True, hideScale = True, rotation = [0,0,0], translation = [0,0,0], scale = [1,1,1], name = None):
	"""
	Constrains theObject to a controller.
	
	controlType - the type of control to be created
	toParent - should the controller's group be constrained to theObject's parent?
	toSecondParent - as before, but constraint to the parent of the parent
	skip - passed as the skip argument when creating an orient constraint from the controller to theObject.
	drive - should theObject be constrainted to the controller?
	hideTranslate - should the translate attributes on the controller be hidden?
	hidScale - should the scale attributes on the controller be hidden?
	rotaton - rotation offset for the controller curve
	translation - translation offset for the controller curve
	scale - scale offset for the controller curve
	name - name of the controller curve. If None, named whatever maya assigns it.
	
	returns - the controller, the controller's group	
	
	"""	
	
	#get a controller and group it
	controller = getController(controlType)
	ctrlGroup = pc.group(controller)
	
	#try to name them
	if name:
		pc.rename(controller, name)
		pc.rename(ctrlGroup, name + "_group")
	
	#we have to use xform instead of setMatrix because the default pymel version that comes with
	#Maya 2011 (1.0.0) errors out on setMatrix... *sigh*
	#ctrlGroup.setMatrix(theObject.getMatrix(ws=True), ws=True) #align group to object
	pc.xform(ctrlGroup, m =theObject.getMatrix(ws=True), ws=True) #align group to object
	controller.t.set(translation)
	controller.r.set(rotation)
	controller.s.set(scale)
	controller.setRotatePivot(theObject.getTranslation("world"), "world")
	#setScalePivot is bugged in Pymel and can't be called, so we just copy the rotate pivot manually
	controller.scalePivot.set(controller.rotatePivot.get())
	pc.makeIdentity(controller, apply = True)
	
	#Constrain the object to the controller if requested
	if(drive):
		constraint = pc.orientConstraint(controller, theObject, mo = True, skip = skip)
		constraint.attr("lockOutput").set(True)
		constraint.attr("interpType").set(2) #shortest interpolation type
	
	#constain the controller's group to a parent if requested
	if(toParent):
		theParent = theObject.getParent()
		if(toSecondParent):
			theParent = theParent.getParent()
		if(theParent):
			parentConstraint = pc.parentConstraint(theParent, ctrlGroup, mo = True)
			parentConstraint.attr("lockOutput").set(True)
	
	#hide requested attributes
	if(hideTranslate):
		pc.setAttr(controller.tx, k = False, l = True, cb = False)
		pc.setAttr(controller.ty, k = False, l = True, cb = False)
		pc.setAttr(controller.tz, k = False, l = True, cb = False)
	if(hideScale):
		pc.setAttr(controller.sx, k = False, l = True, cb = False)
		pc.setAttr(controller.sy, k = False, l = True, cb = False)
		pc.setAttr(controller.sz, k = False, l = True, cb = False)
			
	
	return controller, ctrlGroup

def mirrorJoints(behavior = False, mirrorSide = kMirrorXY, changeSide = True):
	"""
	Wrapper for Maya's mirrorJoint command. Mirrors the selected joint chains.
	If multiple chains are selected, does them one at a time.
	
	behavior - the value for the mirrorBehavior argument
	mirrorSide - Which axis to mirror across
	changeSide - if true, the "side" attribute of the new joints
		is swapped.
	
	Deselects on completion.
	
	"""
	
	sel = pc.selected()
	
	for item in sel:
	
		if mirrorSide == kMirrorXY:
			ret = pc.mirrorJoint(item, mb = behavior, mxy = True)
		elif mirrorSide == kMirrorXZ:
			ret = pc.mirrorJoint(item, mb = behavior, mxz = True)
		else:
			ret = pc.mirrorJoint(item, mb = behavior, myz = True)
	
		if changeSide:
			#go through returned items and switch side of joints
			for name in ret:
				joint = pc.PyNode(name)
				if pc.nodeType(joint) == "joint":
					if joint.side.get() == kSideRight:
						joint.side.set(kSideLeft)
					elif joint.side.get() == kSideLeft:
						joint.side.set(kSideRight)
	
	pc.select(cl = True)
	
def assignColorOverride(node, side = kSideCenter):
	"""
	Turns on the display override for node's shape and
	sets its color based on side.
	
	"""
	
	shape = node.getShape()
	
	if shape:
		shape.overrideEnabled.set(True)
		if side == kSideRight:
			shape.ovc.set(4)
		elif side == kSideLeft:
			shape.ovc.set(6)
		else:
			shape.ovc.set(17)

def getController(controlType):
	"""
	Returns a curve node based on controlType.
	
	"""
	
	if controlType == kCircleCurve:
		return circleController()
	elif controlType == kBoxCurve:
		return boxController()
	elif controlType == kSpatulaCurve:
		return spatulaController()
	elif controlType == kHCurve:
		return hController()
	elif controlType == kCogCurve:
		return cogController()
	elif controlType == kMasterCurve:
		return masterController()
	else:
		return sphereController()
	
def cogController():
	"""
	Returns a nurbs circle of radius 2.5 aligned to the y-axis.
	The points in the cirlce are purubred alternating up and down
	
	"""
	
	curve = pc.circle(nr = [0, 1, 0], ch = True, r = 2.5)[0] #align circle down y-axis
	#offset the cvs so it looks different
	for i in range(8):
		cv = curve.cv[i]
		vector = [0,.5,0]
		if i % 2 != 0:
			vector = [0, -.5, 0]
		cv.translateBy(vector, "object")
	curve.v.setKeyable(False)
	return curve

def masterController():
	"""
	Returns a nurbs circle of radius 5, aligned to the y-axis and set to linear degree
	
	"""
	
	curve = pc.circle(nr = [0, 1, 0], ch = True, d = 1, r = 5)[0] #align circle down y-axis, linear
	curve.v.setKeyable(False)
	return curve

def circleController():
	"""
	Returns a nurbs circle of radius 2.5 aligned to the x-axis.
	
	"""
	
	curve = pc.circle(nr = [1, 0, 0], ch = True, r = 2.5)[0] #align circle down x-axis
	curve.v.setKeyable(False)
	return curve

def boxController():
	"""
	Returns a nurbs curve in the shape of a cube.
	
	"""
	
	curve = pc.curve(d = 1, p = [(-1, -1, -1), (1, -1, -1), \
							(1, -1, 1), (-1, -1, 1), (-1, -1, -1),\
							(-1, 1, -1), (1, 1, -1), (1, -1, -1), \
							(1, -1, 1), (1, 1, 1), (-1, 1, 1), \
							(-1, -1, 1), (-1, 1, 1), (1, 1, 1), \
							(1, 1, -1), (-1, 1, -1), (-1, 1, 1)])
	curve.v.setKeyable(False)
	return curve

def hController():
	"""
	Returns a nurbs curve in the shape of a letter 'H'.
	
	"""
	
	curve = pc.curve(d = 1, p = [(1, 0, 1), (.5, 0, 1), \
								(.5, 0, .20), (-.5, 0, .20), (-.5, 0, 1),\
								(-1, 0, 1), (-1, 0, -1), (-.5, 0, -1), \
								(-.5, 0, -.20), (.5, 0, -.20), (.5, 0, -1), \
								(1, 0, -1), (1, 0, 1)])
	curve.v.setKeyable(False)
	return curve

def spatulaController():
	"""
	Returns a nurbs curve in the shape of a 'spatula'.
	That is, a circular top with a stick going down to the pivot.
	
	"""
	
	curve = pc.curve(d = 1, p = [(-0.064000000000000001, 0, -1.0868728166557649),
								(0.064000000000000001, 0, -1.0868728166557649),
								(0.192, 0, -0.95900542646096909),
								(0.192, 0, -0.83113803626617311),
								(0.064000000000000001, 0, -0.70327064607137735),
								(-0.064000000000000001, 0, -0.70327064607137735),
								(-0.192, 0, -0.83113803626617311),
								(-0.192, 0, -0.95900542646096909),
								(-0.064000000000000001, 0, -1.0868728166557649),
								(0.192, 0, -0.83113803626617311),
								(0.064000000000000001, 0, -0.70327064607137735),
								(-0.192, 0, -0.95900542646096909),
								(-0.192, 0, -0.83113803626617311),
								(0.064000000000000001, 0, -1.0868728166557649),
								(0.192, 0, -0.95900542646096909),
								(-0.064000000000000001, 0, -0.70327064607137735),
								(0, 0, -0.70327064607137735),
								(0, 0, -0.030469289372290098),
								(-0.030500888723165816, 0, 0),
								(0, 0, 0.030469289372289765),
								(0.030500888723165816, 0, 0),
								(0, 0, -0.030469289372290098)])
	curve.v.setKeyable(False)
	return curve

def sphereController():
	"""
	Returns a nurbs curve in the shape of a sphere.
	
	"""
	
	curve = pc.curve(d = 1, p = [(9.5961577477086953e-017, -1, -2.5315347264529309e-016),
								(0.25512228824339328, -1, -1.0408340855860843e-016),
								(0.52088735324739588, -0.89180581244561219, -7.2858385991025898e-017),
								(0.70710678118654746, -0.70710678118654746, -4.163336342344337e-017),
								(0.70710678118654757, -0.70710678118654746, 0.1846990312590647),
								(0.63060193748187077, -0.70710678118654746, 0.36939806251812929),
								(0.5, -0.70710678118654757, 0.50000000000000011),
								(0.63167702028445372, -0.52240774992748296, 0.63167702028445394),
								(0.70710678118654746, -0.26120387496374148, 0.70710678118654768),
								(0.70710678118654768, 7.7037197775489434e-034, 0.70710678118654768),
								(0.89180581244561219, 9.2518585385429722e-018, 0.52240774992748307),
								(1, 1.8503717077085944e-017, 0.26120387496374164),
								(1, 2.7755575615628914e-017, 7.4014868308343765e-017),
								(1, 0.26120387496374153, 1.0639637319324416e-016),
								(0.89332620912569949, 0.52240774992748307, 1.2952601953960158e-016),
								(0.70710678118654768, 0.70710678118654768, 1.5265566588595902e-016),
								(0.70710678118654779, 0.70710678118654757, 0.18469903125906492),
								(0.63060193748187088, 0.70710678118654757, 0.36939806251812957),
								(0.50000000000000011, 0.70710678118654768, 0.50000000000000033),
								(0.36832297971554628, 0.8918058124456123, 0.36832297971554651),
								(0.18039870004873243, 0.99999999999999989, 0.18039870004873262),
								(1.7399673366363375e-016, 1, -1.6804838495427943e-017),
								(-0.18039870004873246, 0.99999999999999989, -0.18039870004873243),
								(-0.36832297971554639, 0.8918058124456123, -0.36832297971554628),
								(-0.50000000000000022, 0.70710678118654768, -0.5),
								(-0.63060193748187099, 0.70710678118654768, -0.36939806251812923),
								(-0.70710678118654791, 0.70710678118654768, -0.18469903125906453),
								(-0.70710678118654791, 0.70710678118654768, 1.6653345369377348e-016),
								(-0.89332620912569971, 0.52240774992748307, 1.8503717077085943e-016),
								(-1.0000000000000004, 0.26120387496374153, 2.1741867565575984e-016),
								(-1.0000000000000004, 2.7755575615628914e-017, 2.2435756955966707e-016),
								(-1.0000000000000002, 4.6259292692714858e-017, -0.26120387496374114),
								(-0.8918058124456123, 6.4763009769800802e-017, -0.52240774992748262),
								(-0.70710678118654768, 7.4014868308343778e-017, -0.70710678118654735),
								(-0.70710678118654768, -0.26120387496374142, -0.70710678118654735),
								(-0.63167702028445405, -0.52240774992748296, -0.63167702028445372),
								(-0.50000000000000022, -0.70710678118654757, -0.49999999999999994),
								(-0.63060193748187077, -0.70710678118654746, -0.36939806251812901),
								(-0.70710678118654768, -0.70710678118654746, -0.18469903125906442),
								(-0.70710678118654768, -0.70710678118654746, 1.9428902930940239e-016),
								(-0.5208873532473961, -0.89180581244561219, 1.2836953722228372e-016),
								(-0.25512228824339334, -1, 6.2450045135165055e-017),
								(9.5961577477086953e-017, -1, -2.5315347264529309e-016),
								(-0.18039870004873232, -1, 0.18039870004873243),
								(-0.36832297971554617, -0.89180581244561219, 0.36832297971554628),
								(-0.5, -0.70710678118654746, 0.50000000000000011),
								(-0.63060193748187077, -0.70710678118654746, 0.36939806251812946),
								(-0.70710678118654768, -0.70710678118654746, 0.18469903125906481),
								(-0.70710678118654791, -0.70710678118654757, 1.7058114180438605e-016),
								(-0.8933262091256996, -0.52240774992748296, 2.1279274638648835e-016),
								(-1.0000000000000002, -0.26120387496374148, 2.312964634635743e-016),
								(-1.0000000000000002, 3.2381504884900401e-017, 2.7755575615628914e-016),
								(-1.0000000000000002, 1.8503717077085941e-017, 0.2612038749637417),
								(-0.8918058124456123, 9.2518585385429707e-018, 0.52240774992748318),
								(-0.70710678118654768, 2.7755575615628914e-017, 0.70710678118654779),
								(-0.70710678118654768, 0.26120387496374153, 0.70710678118654791),
								(-0.63167702028445394, 0.52240774992748307, 0.63167702028445416),
								(-0.50000000000000011, 0.70710678118654768, 0.50000000000000022),
								(-0.63060193748187099, 0.70710678118654768, 0.36939806251812957),
								(-0.70710678118654791, 0.70710678118654768, 0.18469903125906487),
								(-0.70710678118654779, 0.70710678118654768, 1.4282556618875713e-016),
								(-0.5208873532473961, 0.8918058124456123, 1.0061396160665481e-016),
								(-0.25512228824339345, 0.99999999999999989, 4.8572257327350599e-017),
								(1.7399673366363375e-016, 1, -1.6804838495427943e-017),
								(-0.18039870004873254, 0.99999999999999989 ,0.18039870004873251),
								(-0.36832297971554639, 0.8918058124456123, 0.36832297971554645),
								(-0.50000000000000022, 0.70710678118654768, 0.50000000000000033),
								(-0.36939806251812934, 0.70710678118654768, 0.63060193748187099),
								(-0.18469903125906462, 0.70710678118654768, 0.70710678118654791),
								(5.5511151231257827e-017, 0.70710678118654768, 0.70710678118654791),
								(8.3266726846886753e-017, 0.52240774992748307, 0.89332620912569971),
								(1.1102230246251568e-016, 0.26120387496374148, 1.0000000000000004),
								(1.2490009027033011e-016, -2.7755575615628914e-017, 1.0000000000000004),
								(-0.26120387496374126, -1.8503717077085944e-017, 1.0000000000000002),
								(-0.52240774992748285, -9.2518585385429722e-018, 0.8918058124456123),
								(-0.70710678118654757, -7.7037197775489434e-034, 0.70710678118654768),
								(-0.70710678118654757, -0.26120387496374148, 0.70710678118654768),
								(-0.63167702028445383, -0.52240774992748296, 0.63167702028445394),
								(-0.5, -0.70710678118654757, 0.50000000000000011),
								(-0.36939806251812918, -0.70710678118654746, 0.63060193748187088),
								(-0.18469903125906448, -0.70710678118654746, 0.70710678118654768),
								(1.3877787807814457e-016, -0.70710678118654746, 0.70710678118654768),
								(1.1796119636642288e-016, -0.89180581244561219, 0.52088735324739599),
								(9.7144514654701197e-017, -1, 0.25512228824339328),
								(9.5961577477086953e-017, -1, -2.5315347264529309e-016),
								(0.18039870004873226, -1, -0.1803987000487326),
								(0.36832297971554601, -0.89180581244561219, -0.36832297971554656),
								(0.49999999999999978, -0.70710678118654746, -0.50000000000000044),
								(0.63060193748187054, -0.70710678118654746, -0.36939806251812968),
								(0.70710678118654746, -0.70710678118654746, -0.18469903125906478),
								(0.70710678118654757, -0.70710678118654757, -3.6429192995512949e-017),
								(0.89332620912569927, -0.52240774992748296, 0),
								(1, -0.26120387496374148, 4.163336342344337e-017),
								(1, 3.2381504884900401e-017, 2.7755575615628914e-017),
								(1, 4.6259292692714858e-017, -0.26120387496374159),
								(0.89180581244561208, 6.4763009769800802e-017, -0.5224077499274834),
								(0.70710678118654724, 8.3266726846886741e-017, -0.70710678118654813),
								(0.70710678118654724, 0.26120387496374159, -0.70710678118654813),
								(0.63167702028445372, 0.52240774992748307, -0.63167702028445438),
								(0.49999999999999994, 0.70710678118654768, -0.50000000000000033),
								(0.63060193748187077, 0.70710678118654768, -0.36939806251812951),
								(0.70710678118654768, 0.70710678118654768, -0.18469903125906462),
								(0.70710678118654779, 0.70710678118654768, 1.393561192368035e-016),
								(0.5208873532473961, 0.8918058124456123, 1.4918621893400541e-016),
								(0.25512228824339339, 0.99999999999999989, 1.457167719820518e-016),
								(1.7399673366363375e-016, 1, -1.6804838495427943e-017),
								(0.18039870004873249, 0.99999999999999989, -0.18039870004873251),
								(0.36832297971554628, 0.8918058124456123, -0.36832297971554656),
								(0.5, 0.70710678118654768, -0.50000000000000044),
								(0.36939806251812912, 0.70710678118654768, -0.63060193748187121),
								(0.18469903125906442, 0.70710678118654768, -0.70710678118654791),
								(-2.2204460492503131e-016, 0.70710678118654768, -0.70710678118654791),
								(-3.0531133177191805e-016, 0.52240774992748307, -0.89332620912569971),
								(-3.6082248300317592e-016, 0.26120387496374159, -1.0000000000000004),
								(-4.163336342344337e-016, 8.3266726846886741e-017, -1.0000000000000002),
								(0.26120387496374098, 8.3266726846886741e-017, -1.0000000000000004),
								(0.5224077499274824, 8.3266726846886741e-017, -0.89180581244561274),
								(0.70710678118654724, 8.3266726846886741e-017, -0.70710678118654813),
								(0.70710678118654713, -0.26120387496374142, -0.70710678118654813),
								(0.63167702028445349, -0.52240774992748296, -0.63167702028445438),
								(0.49999999999999978, -0.70710678118654746, -0.50000000000000044),
								(0.36939806251812896, -0.70710678118654746, -0.63060193748187121),
								(0.18469903125906428, -0.70710678118654746, -0.70710678118654779),
								(-3.1918911957973251e-016, -0.70710678118654746, -0.70710678118654768),
								(-2.6367796834847468e-016, -0.89180581244561219, -0.5208873532473961),
								(-1.6653345369377348e-016, -1, -0.25512228824339339),
								(9.5961577477086953e-017, -1, -2.5315347264529309e-016),
								(-0.18039870004873251, -1, -0.18039870004873237),
								(-0.36832297971554639, -0.89180581244561219, -0.36832297971554617),
								(-0.50000000000000011, -0.70710678118654746, -0.49999999999999978),
								(-0.36939806251812946, -0.70710678118654746, -0.63060193748187054),
								(-0.18469903125906492, -0.70710678118654746, -0.70710678118654746),
								(-3.2150208421436826e-016, -0.70710678118654757, -0.70710678118654791),
								(-3.7932620008026185e-016, -0.52240774992748296, -0.8933262091256996),
								(-3.9782991715734778e-016, -0.26120387496374142, -1.0000000000000002),
								(-3.8857805861880479e-016, 8.3266726846886741e-017, -1.0000000000000002),
								(-0.26120387496374176, 8.3266726846886741e-017, -1),
								(-0.52240774992748307, 8.3266726846886741e-017, -0.89180581244561208),
								(-0.70710678118654768, 8.3266726846886741e-017, -0.70710678118654735),
								(-0.70710678118654779, 0.26120387496374159, -0.70710678118654746),
								(-0.63167702028445405, 0.52240774992748307, -0.63167702028445383),
								(-0.50000000000000022, 0.70710678118654768, -0.5),
								(-0.36939806251812951, 0.70710678118654768, -0.63060193748187088),
								(-0.18469903125906489, 0.70710678118654768, -0.70710678118654768),
								(-2.2204460492503131e-016, 0.70710678118654768, -0.70710678118654791),
								(-1.3877787807814457e-016, 0.8918058124456123, -0.52088735324739621),
								(-2.7755575615628914e-017, 0.99999999999999989, -0.25512228824339339),
								(1.7399673366363375e-016, 1, -1.6804838495427943e-017),
								(-3.4694469519536142e-017, 0.99999999999999989, 0.2551222882433935),
								(1.0408340855860843e-017, 0.8918058124456123, 0.52088735324739621),
								(4.6837533851373792e-017, 0.70710678118654768, 0.70710678118654791),
								(0.18469903125906473, 0.70710678118654757, 0.70710678118654791),
								(0.36939806251812934, 0.70710678118654757, 0.63060193748187099),
								(0.50000000000000011, 0.70710678118654768, 0.50000000000000033),
								(0.63167702028445394, 0.52240774992748307, 0.63167702028445416),
								(0.70710678118654757, 0.26120387496374153, 0.70710678118654791),
								(0.70710678118654746, 2.7755575615628914e-017, 0.70710678118654779),
								(0.52240774992748296, -9.2518585385429722e-018, 0.8918058124456123),
								(0.26120387496374159, -1.8503717077085944e-017, 1),
								(1.6653345369377348e-016, -1.8503717077085944e-017, 1),
								(1.3877787807814457e-016, -0.26120387496374153, 1.0000000000000002),
								(1.3877787807814457e-016, -0.52240774992748296, 0.89332620912569949),
								(1.2836953722228372e-016, -0.70710678118654757, 0.70710678118654768),
								(0.18469903125906476, -0.70710678118654746, 0.70710678118654757),
								(0.36939806251812929, -0.70710678118654746, 0.63060193748187077),
								(0.5, -0.70710678118654746, 0.5),
								(0.36832297971554628, -0.89180581244561219, 0.36832297971554623),
								(0.18039870004873243, -1, 0.18039870004873235),
								(9.5961577477086953e-017, -1, -2.5315347264529309e-016)])
	curve.v.setKeyable(False)
	return curve


def pyWarning( warningString ):
	""" print a warning message """
	import maya.mel as mel
	try: 
		mel.eval('warning "%s"' % warningString)
	except: 
		pass

def pyError( errorString ):
	""" print an error message """
	import maya.mel as mel
	try: 
		mel.eval('error "%s"' % errorString)
	except: 
		pass
	
	
#####################################################
##### UI Code #######################################
#####################################################

_kWindowName = "eRigUtilsWindowName" #name of the window object
_kWindowWidth = 300 #dimensions of the window
_kWindowHeight = 400
_kWindowTitle = "eRig" #displayed title of the window

class MainWindow(object):
	"""Class for creating the eRig window"""
	
	def __init__(self):
		"""Builds the eRig window, but doesn't show it"""
		
		#check if the window already exists
		if (pc.window(_kWindowName, q = True, ex = True)):
			pc.deleteUI(_kWindowName)
		
		#see if a windowPref exists for the window and create one if not
		if not (pc.windowPref(_kWindowName, ex = True)):
			pc.windowPref(_kWindowName)
		
		#set the window prefs
		pc.windowPref(_kWindowName, e = True, widthHeight = \
					[_kWindowWidth, _kWindowHeight])
		
		#create the window
		pc.window(_kWindowName, t = _kWindowTitle, mxb = False, \
				iconName = _kWindowTitle)
		
		#make a form layout to fix the UI
		self.form = pc.formLayout()
		
		#make a tab layout for the window, then create the tabs
		self.tabLayout = pc.tabLayout(cr=True)
		pc.formLayout( self.form, edit=True, attachForm=((self.tabLayout, 'top', 0),\
												(self.tabLayout, 'left', 0),\
												(self.tabLayout, 'bottom', 0),\
												(self.tabLayout, 'right', 0)) )
				
		self.buildFirstTab()
		pc.setParent(self.tabLayout)
		
		self.buildSecondTab()
		pc.setParent(self.tabLayout)
		
		self.buildThirdTab()
		pc.setParent(self.tabLayout)
		
		pc.tabLayout(self.tabLayout, edit=True, \
					tabLabel=((self.firstTab, 'Biped'),\
							(self.secondTab, 'Parts'),\
							(self.thirdTab, 'Utils')))
		
	
	def buildFirstTab(self):
		"""Builds the biped tab"""
		
		self.firstTab = pc.columnLayout(adj = True, cal = "center")
		
		#add a row layout for front axis selection
		pc.rowLayout(nc = 3, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
			columnWidth3=(60,40,40))
		pc.text(label="Front Axis:")
		pc.radioCollection()
		pc.radioButton(label="X", onc=partial(self.setBipedFrontAxisCallback, True))
		pc.radioButton(label="Z", sl = True, onc=partial(self.setBipedFrontAxisCallback, False))
		
		self.frontBipedAxis = False
		
		#back up to the column layout
		pc.setParent("..")
		pc.separator(h = 5)
		
		pc.text("Biped Options:")
		pc.separator(h = 5, style = "none")
		
		
		#add a row layout for extras
		pc.rowLayout(nc = 3, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
			columnWidth3=(70,70, 130))		
		self.eyesCheckBox = pc.checkBox(label="Add Eyes")
		self.jawCheckBox = pc.checkBox(label="Add Jaw")
		self.bipedProxyCheckBox = pc.checkBox(label="Create Proxy Geometry", v = True)
		
		#back up to the column layout
		pc.setParent("..")
		pc.separator(h = 5, style = "none")
		
		
		#add a row layout for arm options
		pc.rowLayout(nc = 4, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0), (4,'both',0)],
			columnWidth4=(40,45,45,45))
		pc.text(label="Arms:")
		pc.radioCollection()
		pc.radioButton(label="FK", onc=partial(self.setBipedArmFkIk, kFkOnly))
		pc.radioButton(label="IK", onc=partial(self.setBipedArmFkIk, kIkOnly))
		pc.radioButton(label="FK/IK", sl = True, onc=partial(self.setBipedArmFkIk, kFkIk))
		
		self.bipedArmFkIk = 2
		
		#back up to the column layout
		pc.setParent("..")
		pc.separator(h = 5, style = "none")
		
		#add a row layout for legs options
		pc.rowLayout(nc = 4, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0), (4,'both',0)],
			columnWidth4=(40,45,45,45))
		pc.text(label="Legs:")
		pc.radioCollection()
		pc.radioButton(label="FK", onc=partial(self.setBipedLegFkIk, kFkOnly))
		pc.radioButton(label="IK", onc=partial(self.setBipedLegFkIk, kIkOnly))
		pc.radioButton(label="FK/IK", sl = True, onc=partial(self.setBipedLegFkIk, kFkIk))
		
		self.bipedLegFkIk = 2
		
		#back up to the column layout
		pc.setParent("..")
		pc.separator(h = 5, style = "none")
		
		#make a slider for number of spine joints
		self.numBipedSpines = pc.intSliderGrp(field = True, label = "Num Spine Joints: ", minValue = 1, maxValue = 10,\
					fieldMinValue = 1, fieldMaxValue = 30, value = 3, columnAlign=(1, 'left'),
					columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
					columnWidth3=(90,45,45), adj = 3)
		
		pc.separator(h = 5, style = "none")
		
		#make a slider for number of neck joints
		self.numBipedNecks = pc.intSliderGrp(field = True, label = "Num Neck Joints: ", minValue = 1, maxValue = 10,\
					fieldMinValue = 1, fieldMaxValue = 30, value = 1, columnAlign=(1, 'left'),
					columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
					columnWidth3=(90,45,45), adj = 3)
		
		pc.separator(h = 5, style = "none")
		
		pc.button(label="Generate Biped Skeleton", c = self.createBipedSkeleton)
		pc.button(label="Build Rig (select COG)", c = self.buildRig)
	
	def buildSecondTab(self):
		"""Builds the parts tab"""
		
		self.secondTab = pc.formLayout()
		
		#make a tab layout for the window, then create the tabs
		scrollLayout = pc.scrollLayout(cr = True, hst = 0)
		pc.formLayout( self.secondTab, edit=True, attachForm=((scrollLayout, 'top', 0),\
												(scrollLayout, 'left', 0),\
												(scrollLayout, 'bottom', 0),\
												(scrollLayout, 'right', 0)) )
		mainLayout = pc.columnLayout(adj = True, cal = "center")
		
		##
		pc.frameLayout(l = "COG", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
				
		#add a row layout for front axis selection
		self.partsFrontAxis = False
		
		def partsFrontAxisCB(value, *args):
			self.partsFrontAxis = value
		
		pc.rowLayout(nc = 3, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
			columnWidth3=(60,40,40))
		pc.text(label="Front Axis:")
		pc.radioCollection()
		pc.radioButton(label="X", onc=partial(partsFrontAxisCB, True))
		pc.radioButton(label="Z", sl = True, onc=partial(partsFrontAxisCB, False))
		
		#back up to the column layout
		pc.setParent("..")
		
		pc.columnLayout(adj = True, cal = "left")
		self.partsProxyCheckBox = pc.checkBox(label="Create Proxy Geometry", v = True)
		
		#back up to the column layout
		pc.setParent("..")
		
		#and a button
		pc.button(label="Create COG", c = self.createCog)
		
		pc.setParent(mainLayout)
		
		
		##
		pc.frameLayout(l = "Spine", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
		
		#make a slider for number of spine joints
		self.numPartsSpines = pc.intSliderGrp(field = True, label = "Num Spine Joints: ", minValue = 1, maxValue = 10,\
					fieldMinValue = 1, fieldMaxValue = 30, value = 3, columnAlign=(1, 'left'),
					columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
					columnWidth3=(90,45,45), adj = 3)
		
		#add a row layout for side selection
		self.partsSpineSide = kSideCenter
		
		def partsSpineSideCB(value, *args):
			self.partsSpineSide = value
		
		pc.rowLayout(nc = 4, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0), (4,'both',0)],
			columnWidth4=(40,60,60, 60))
		pc.text(label="Side:")
		pc.radioCollection()
		pc.radioButton(label="Right", onc=partial(partsSpineSideCB, kSideRight))
		pc.radioButton(label="Center", sl = True, onc=partial(partsSpineSideCB, kSideCenter))
		pc.radioButton(label="Left", onc=partial(partsSpineSideCB, kSideLeft))
		
		#back up to the column layout
		pc.setParent("..")
		

		#add name field
		self.spinePartsName = pc.textFieldGrp(adj = 2, l = 'Name:',
											columnAlign=(1, 'left'),
											columnAttach=[(1,'both',0), (2,'both',0)],
											columnWidth2=(35,60))
		
		
		#and a button
		pc.button(label="Create Spine", c = self.createSpine)
		
		pc.setParent(mainLayout)
		
		
		##
		pc.frameLayout(l = "Hips", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
		
		pc.button(label="Create Hips", c = self.createHips)
		
		pc.setParent(mainLayout)
		
		
		##
		pc.frameLayout(l = "Leg", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
		
		#add a row layout for side selection
		self.partsLegSide = kSideRight
		
		def partsLegSideCB(value, *args):
			self.partsLegSide = value
		
		pc.rowLayout(nc = 3, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
			columnWidth3=(40,60,60))
		pc.text(label="Side:")
		pc.radioCollection()
		pc.radioButton(label="Right", sl = True, onc=partial(partsLegSideCB, kSideRight))
		pc.radioButton(label="Left", onc=partial(partsLegSideCB, kSideLeft))
		
		#back up to the column layout
		pc.setParent("..")
		
		#add a row layout for ik/fk selection
		self.partsLegFkIk = kFkIk
		
		def partsLegFkIkCB(value, *args):
			self.partsLegFkIk = value
		
		pc.rowLayout(nc = 4, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0), (4,'both',0)],
			columnWidth4=(40,45,45,45))
		pc.text(label="Type:")
		pc.radioCollection()
		pc.radioButton(label="FK", onc=partial(partsLegFkIkCB, kFkOnly))
		pc.radioButton(label="IK", onc=partial(partsLegFkIkCB, kIkOnly))
		pc.radioButton(label="FK/IK", sl = True, onc=partial(partsLegFkIkCB, kFkIk))
		
		#back up to the column layout
		pc.setParent("..")
		
		#and a button
		pc.button(label="Create Leg", c = self.createLeg)
		
		pc.setParent(mainLayout)
		
		
		##
		pc.frameLayout(l = "Clavicle", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
		
		#add a row layout for side selection
		self.partsClavSide = kSideRight
		
		def partsClavSideCB(value, *args):
			self.partsClavSide = value
		
		pc.rowLayout(nc = 3, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
			columnWidth3=(40,60,60))
		pc.text(label="Side:")
		pc.radioCollection()
		pc.radioButton(label="Right", sl = True, onc=partial(partsClavSideCB, kSideRight))
		pc.radioButton(label="Left", onc=partial(partsClavSideCB, kSideLeft))
		
		#back up to the column layout
		pc.setParent("..")
		
		#and a button
		pc.button(label="Create Clavicle", c = self.createClav)
		
		pc.setParent(mainLayout)
		
		
		##
		pc.frameLayout(l = "Arm", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
		
		#add a row layout for front axis selection
		self.partsArmSide = kSideRight
		
		def partsArmSideCB(value, *args):
			self.partsArmSide = value
		
		pc.rowLayout(nc = 3, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
			columnWidth3=(40,60,60))
		pc.text(label="Side:")
		pc.radioCollection()
		pc.radioButton(label="Right", sl = True, onc=partial(partsArmSideCB, kSideRight))
		pc.radioButton(label="Left", onc=partial(partsArmSideCB, kSideLeft))
		
		#back up to the column layout
		pc.setParent("..")
		
		#add a row layout for ik/fk selection
		self.partsArmFkIk = kFkIk
		
		def partsArmFkIkCB(value, *args):
			self.partsArmFkIk = value
		
		pc.rowLayout(nc = 4, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0), (4,'both',0)],
			columnWidth4=(40,45,45,45))
		pc.text(label="Type:")
		pc.radioCollection()
		pc.radioButton(label="FK", onc=partial(partsArmFkIkCB, kFkOnly))
		pc.radioButton(label="IK", onc=partial(partsArmFkIkCB, kIkOnly))
		pc.radioButton(label="FK/IK", sl = True, onc=partial(partsArmFkIkCB, kFkIk))
		
		#back up to the column layout
		pc.setParent("..")
		
		#and a button
		pc.button(label="Create Arm", c = self.createArm)
		pc.setParent(mainLayout)
		
		
		##
		pc.frameLayout(l = "Neck/Head", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
		
		#make a slider for number of neck joints
		self.numPartsNecks = pc.intSliderGrp(field = True, label = "Num Neck Joints: ", minValue = 1, maxValue = 10,\
					fieldMinValue = 1, fieldMaxValue = 30, value = 1, columnAlign=(1, 'left'),
					columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
					columnWidth3=(90,45,45), adj = 3)
		
		#and a button
		pc.button(label="Create Head/Neck", c = self.createHead)
		
		pc.setParent(mainLayout)
		
		
		##
		pc.frameLayout(l = "Eyes", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
		
		#add a row layout for side selection
		self.partsEyeSide = kSideRight
		
		def partsEyeSideCB(value, *args):
			self.partsEyeSide = value
		
		pc.rowLayout(nc = 3, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0)],
			columnWidth3=(40,60,60))
		pc.text(label="Side:")
		pc.radioCollection()
		pc.radioButton(label="Right", sl = True, onc=partial(partsEyeSideCB, kSideRight))
		pc.radioButton(label="Left", onc=partial(partsEyeSideCB, kSideLeft))
		
		#back up to the column layout
		pc.setParent("..")
		
		#and a button
		pc.button(label="Create Eye", c = self.createEye)
		
		pc.setParent(mainLayout)
		
		
		##
		pc.frameLayout(l = "Jaw", cl = True, cll = True, bs = "etchedIn")
		pc.columnLayout(adj = True, cal = "center")
		
		pc.button(label="Create Jaw", c = self.createJaw)
		
		pc.setParent(mainLayout)
		
	def buildThirdTab(self):
		"""Builds the utils tab"""
		
		self.thirdTab = pc.columnLayout(adj = True, cal = "center")
		
		
		#add a row layout for mirror axis
		self.utilsMirrorAxis = kMirrorYZ
		
		def utilsMirrorAxisCB(value, *args):
			self.utilsMirrorAxis = value
		
		pc.rowLayout(nc = 4, columnAlign=(1, 'left'),
			columnAttach=[(1,'both',0), (2,'both',0), (3,'both',0), (4,'both',0)],
			columnWidth4=(65,40,40,40))
		pc.text(label="Mirror Axis:")
		pc.radioCollection()
		pc.radioButton(label="X", sl = True, onc=partial(utilsMirrorAxisCB, kMirrorYZ))
		pc.radioButton(label="Y", onc=partial(utilsMirrorAxisCB, kMirrorXZ))
		pc.radioButton(label="Z", onc=partial(utilsMirrorAxisCB, kMirrorXY))
		
		#back up to the column layout
		pc.setParent("..")
		
		pc.columnLayout(adj = True, cal = "left")
		self.utilsMirrorSide = pc.checkBox(label="Change Side", v = True)
		pc.setParent("..")
		
		#and a button
		pc.button(label="Mirror Selected Joints", c = self.mirrorJoints)
		
		#add a seperator
		pc.separator(h = 5)
		
		pc.text("Skinning Utils:")
		pc.separator(h = 5, style = "none")
		pc.button(label="Lock Weights For Hierarchy", c = self.lockWeights)
		pc.button(label="Unlock Weights For Hierarchy", c = self.unlockWeights)
		
	def lockWeights(self, *args):
		"""Locks skin wieghts for the selected joints and their children"""
		selection = pc.selected()
		for node in selection:
			holdWeights(True, node)
	
	def unlockWeights(self, *args):
		"""Unlocks skin wieghts for the selected joints and their children"""
		selection = pc.selected()
		for node in selection:
			holdWeights(False, node)
	
	def mirrorJoints(self, *args):
		"""Attempts to mirror all selected joints, and groups the action into an undo chunk"""
		try:
			pc.undoInfo(ock = True)
			mirrorJoints(True, self.utilsMirrorAxis, self.utilsMirrorSide.getValue())
		finally:
			pc.undoInfo(cck = True)
	
	###calbacks
	def setBipedFrontAxisCallback(self, axis, *args):
		self.frontBipedAxis = axis
	
	def setBipedArmFkIk(self, selection, *args):
		self.bipedArmFkIk = selection
	
	def setBipedLegFkIk(self, selection, *args):
		self.bipedLegFkIk = selection
	
	def setValueCallback(self, variable, value, *args):
		variable = value
	
	def createBipedSkeleton(self, *args):
		"""
		Creates a generic base biped setup based on the user's selected options
		Creates an undo chunck.
		
		"""
		
		#first grab values
		useEyes = self.eyesCheckBox.getValue()
		useJaw = self.jawCheckBox.getValue()
		createProxy = self.bipedProxyCheckBox.getValue()
		
		numSpine = self.numBipedSpines.getValue()
		numNeck = self.numBipedNecks.getValue()
		
		#now make the skeleton
		#wrap in tye/finally so that we can open an undo chunk and make sure it closes even if there's an error
		try:
			pc.undoInfo(ock = True)
			buildDefaultSkeleton(useEyes = useEyes, useJaw = useJaw, numNecks = numNeck, numSpines = numSpine,\
								xFrontAxis = self.frontBipedAxis, armFkIk = self.bipedArmFkIk, legFkIk = self.bipedLegFkIk,\
								createProxy = createProxy)
		finally:
			pc.undoInfo(cck = True)
	
	def buildRig(self, *args):
		"""
		Builds the rig based on the selected Cog
		Opens an undo chunck
		
		"""
		
		#wrap in try/finally so that we can open an undo chunk and make sure it closes even if there's an error
		try:
			pc.undoInfo(ock = True)
			buildRig()
		finally:
			pc.undoInfo(cck = True)
	
	def show(self):
		"""Shows this window"""
		pc.showWindow(_kWindowName)
	
	##Methods to create each component type
	def createCog(self, *args):
		try:
			pc.undoInfo(ock = True)
			createCog(self.partsFrontAxis, self.partsProxyCheckBox.getValue())
		finally:
			pc.undoInfo(cck = True)
	
	def createSpine(self, *args):
		try:
			pc.undoInfo(ock = True)
			createSpine(self.numPartsSpines.getValue(), self.partsSpineSide, self.spinePartsName.getText())
		finally:
			pc.undoInfo(cck = True)
	
	def createHips(self, *args):
		try:
			pc.undoInfo(ock = True)
			createHips()
		finally:
			pc.undoInfo(cck = True)
	
	def createLeg(self, *args):
		try:
			pc.undoInfo(ock = True)
			createLeg(self.partsLegSide, self.partsLegFkIk)
		finally:
			pc.undoInfo(cck = True)
	
	def createArm(self, *args):
		try:
			pc.undoInfo(ock = True)
			createArm(self.partsArmSide, self.partsArmFkIk)
		finally:
			pc.undoInfo(cck = True)
	
	def createClav(self, *args):
		try:
			pc.undoInfo(ock = True)
			createClavicle(self.partsClavSide)
		finally:
			pc.undoInfo(cck = True)
	
	def createHead(self, *args):
		try:
			pc.undoInfo(ock = True)
			createHead(self.numPartsNecks.getValue())
		finally:
			pc.undoInfo(cck = True)
	
	def createEye(self, *args):
		try:
			pc.undoInfo(ock = True)
			createEye(self.partsEyeSide)
		finally:
			pc.undoInfo(cck = True)
	
	def createJaw(self, *args):
		try:
			pc.undoInfo(ock = True)
			createJaw()
		finally:
			pc.undoInfo(cck = True)

def holdWeights(setHold, node):
	"""
	Sets the skin weights for node and all children (that are joints)
	to be set to setHold.
													
	"""
	
	if pc.nodeType(node) == "joint":
		if node.hasAttr("lockInfluenceWeights"):
			node.lockInfluenceWeights.set(setHold)
		for child in node.getChildren():
			holdWeights(setHold, child)


def openWindow():
	"""Creates the eRig window and opens it"""
	
	window = MainWindow()
	window.show()



